import numpy as np
from scipy.interpolate import interp1d

from solcore import asUnit, constants
from solcore.state import State
from .ddModel import driftdiffusion as dd
from .DeviceStructure import LoadAbsorption, CreateDeviceStructure, CalculateAbsorptionProfile

Epsi0 = constants.vacuum_permittivity
q = constants.q
pi = constants.pi
h = constants.h
kb = constants.kb
m0 = constants.electron_mass

log = dd.log_file

pdd_options = State()

# Mesh control
pdd_options.meshpoints = -400
pdd_options.growth_rate = 0.7
pdd_options.coarse = 20e-9
pdd_options.fine = 1e-9
pdd_options.ultrafine = 0.2e-9

# Convergence control
pdd_options.clamp = 20
pdd_options.nitermax = 100
pdd_options.ATol = 1e-14
pdd_options.RTol = 1e-6

# Recombination control
pdd_options.srh = 1
pdd_options.rad = 1
pdd_options.aug = 0
pdd_options.sur = 1
pdd_options.gen = 0

# Output control
pdd_options.output_equilibrium = 1
pdd_options.output_sc = 1
pdd_options.output_iv = 1
pdd_options.output_qe = 1


# Functions for creating the streucture in the fortran variables and executing the DD solver
def ProcessStructure(device, meshpoints, wavelengths=None):
    """ This function reads a dictionary containing all the device structure, extract the electrical and optical
    properties of the materials, and loads all that information into the Fortran variables. Finally, it initialises the
    device (in Fortran) calculating an initial mesh and all the properties as a function of the position.

    :param device: A dictionary containing the device structure. See PDD.DeviceStructure
    :param wavelengths: (Optional) Wavelengths at which to calculate the optical properties.
    :return: Dictionary containing the device structure properties as a function of the position.
    """
    print('Processing structure...')
    # First, we clean any previous data from the Fortran code
    dd.reset()
    output = {}

    if wavelengths is not None:
        output['Optics'] = {}
        output['Optics']['wavelengths'] = wavelengths

    # We dump the structure information to the Fotran module and initialise the structure
    i = 0
    while i < device['numlayers']:
        layer = device['layers'][i]['properties']
        args_list = [layer['width'],
                     asUnit(layer['band_gap'], 'eV'),
                     asUnit(layer['electron_affinity'], 'eV'),
                     layer['electron_mobility'],
                     layer['hole_mobility'],
                     layer['Nc'],
                     layer['Nv'],
                     layer['electron_minority_lifetime'],
                     layer['hole_minority_lifetime'],
                     layer['relative_permittivity'] * Epsi0,
                     layer['radiative_recombination'],
                     layer['electron_auger_recombination'],
                     layer['hole_auger_recombination'],
                     layer['Na'],
                     layer['Nd']]
        dd.addlayer(args=args_list)
        i = i + device['layers'][i]['numlayers']

    # We set the surface recombination velocities. This needs to be improved at some point
    # to consider other boundary conditions
    dd.frontboundary("ohmic", device['sn'], device['sp'], 0)
    dd.backboundary("ohmic", device['sn'], device['sp'], 0)

    dd.initdevice(meshpoints)
    print('...done!\n')

    output['Properties'] = DumpInputProperties()
    return output


def equilibrium_pdd(junction, options):
    """ Solves the PDD equations under equilibrium: in the dark with no external current and zero applied voltage.

    :param junction: A junction object
    :param options: Options to be passed to the solver
    :return: None
    """
    T = options.T
    wl = options.wavelength
    output_info = options.output_equilibrium

    SetConvergenceParameters(options)
    SetMeshParameters(options)
    SetRecombinationParameters(options)

    device = CreateDeviceStructure('Junction', T=T, layers=junction)

    print('Solving equilibrium...')
    output = ProcessStructure(device, options.meshpoints, wavelengths=wl)
    dd.gen = 0

    dd.equilibrium(output_info)
    print('...done!\n')

    output['Bandstructure'] = DumpBandStructure()
    junction.equilibrium_data = State(**output)


def short_circuit_pdd(junction, options):
    """ Solves the devices electronic properties at short circuit. Internally, it calls Equilibrium.

    :param junction: A junction object
    :param options: Options to be passed to the solver
    :return: None
    """

    # We run equilibrium
    equilibrium_pdd(junction, options)

    wl = options.wavelength
    wl, ph = options.light_source.spectrum(x=wl, output_units='photon_flux_per_m')

    z = junction.equilibrium_data['Bandstructure']['x']
    absorption = junction.absorbed if hasattr(junction, 'absorbed') else lambda x: 0
    abs_profile = CalculateAbsorptionProfile(z, wl, absorption)

    dd.set_generation(abs_profile)
    dd.gen = 1

    dd.illumination(ph)

    dd.lightsc(options.output_sc, 1)
    print('...done!\n')

    output = {'Bandstructure': DumpBandStructure()}
    junction.short_circuit_data = State(**output)


def iv_pdd(junction, options):
    """ Calculates the IV curve of the device between 0 V and a given voltage. Depending on the options, the IV will be calculated in the dark (calling the equilibrium_pdd function) or under illumination (calling the short_circuit_pdd function). If the voltage range has possitive and negative values, the problem is solved twice: from 0 V to the maximu positive and from 0 V to the maximum negative, concatenating the results afterwards.

    :param junction: A junction object
    :param options: Options to be passed to the solver
    :return: None
    """
    print('Solving IV...')
    light = options.light_iv
    output_info = options.output_iv
    T = options.T

    junction.voltage = options.internal_voltages

    Eg = 10
    for layer in junction:
        Eg = min([Eg, layer.material.band_gap / q])

    s = True if junction[0].material.Na >= junction[0].material.Nd else False

    if s:
        vmax = min(Eg + 3 * kb * T / q, max(junction.voltage))
        vmin = min(junction.voltage)
    else:
        vmax = max(junction.voltage)
        vmin = max(-Eg - 3 * kb * T / q, min(junction.voltage))

    vstep = junction.voltage[1] - junction.voltage[0]

    # POSITIVE RANGE
    output_pos = None
    if vmax > 0:
        if light:
            short_circuit_pdd(junction, options)
        else:
            equilibrium_pdd(junction, options)

        dd.runiv(vmax, vstep, output_info, 0)
        output_pos = {'Bandstructure': DumpBandStructure(), 'IV': DumpIV()}

    # NEGATIVE RANGE
    output_neg = None
    if vmin < 0:
        if light:
            short_circuit_pdd(junction, options)
        else:
            equilibrium_pdd(junction, options)

        dd.runiv(vmin, (-1 * vstep), output_info, 0)
        output_neg = {'Bandstructure': DumpBandStructure(), 'IV': DumpIV()}

    print('...done!\n')

    # Now we need to put together the data for the possitive and negative regions.
    junction.pdd_data = State({'possitive_V': output_pos, 'negative_V': output_neg})

    if output_pos is None:
        V = output_neg['IV']['V']
        J = output_neg['IV']['J']
        Jrad = output_neg['IV']['Jrad']
        Jsrh = output_neg['IV']['Jsrh']
        Jaug = output_neg['IV']['Jaug']
        Jsur = output_neg['IV']['Jsur']
    elif output_neg is None:
        V = output_pos['IV']['V']
        J = output_pos['IV']['J']
        Jrad = output_pos['IV']['Jrad']
        Jsrh = output_pos['IV']['Jsrh']
        Jaug = output_pos['IV']['Jaug']
        Jsur = output_pos['IV']['Jsur']
    else:
        V = np.concatenate((output_neg['IV']['V'][:0:-1], output_pos['IV']['V']))
        J = np.concatenate((output_neg['IV']['J'][:0:-1], output_pos['IV']['J']))
        Jrad = np.concatenate((output_neg['IV']['Jrad'][:0:-1], output_pos['IV']['Jrad']))
        Jsrh = np.concatenate((output_neg['IV']['Jsrh'][:0:-1], output_pos['IV']['Jsrh']))
        Jaug = np.concatenate((output_neg['IV']['Jaug'][:0:-1], output_pos['IV']['Jaug']))
        Jsur = np.concatenate((output_neg['IV']['Jsur'][:0:-1], output_pos['IV']['Jsur']))

    # Finally, we calculate the currents at the desired voltages
    R_shunt = min(junction.R_shunt, 1e14) if hasattr(junction, 'R_shunt') else 1e14

    junction.current = np.interp(junction.voltage, V, J) + junction.voltage / R_shunt
    Jrad = np.interp(junction.voltage, V, Jrad)
    Jsrh = np.interp(junction.voltage, V, Jsrh)
    Jaug = np.interp(junction.voltage, V, Jaug)
    Jsur = np.interp(junction.voltage, V, Jsur)

    junction.iv = interp1d(junction.voltage, junction.current, kind='linear', bounds_error=False, assume_sorted=True,
                           fill_value=(junction.current[0], junction.current[-1]))
    junction.recombination_currents = State({"Jrad": Jrad, "Jsrh": Jsrh, "Jaug": Jaug, "Jsur": Jsur})


def qe_pdd(junction, options):
    """ Calculates the quantum efficiency of the device at short circuit. Internally it calls ShortCircuit

    :param junction: A junction object
    :param options: Options to be passed to the solver
    :return: None
    """
    print('Solving quantum efficiency...')
    output_info = options.output_qe

    short_circuit_pdd(junction, options)

    dd.runiqe(output_info)
    print('...done!\n')

    output = {}
    output['QE'] = DumpQE()
    output['QE']['WL'] = options.wavelength

    z = junction.short_circuit_data['Bandstructure']['x']
    absorbed_per_wl = np.trapz(junction.absorbed(z), z, axis=0)

    # This is redundant but useful to keep the same format than the other solvers
    junction.qe = State(**output['QE'])
    junction.qe['IQE'] = junction.qe['EQE']/np.maximum(absorbed_per_wl, 0.00001)

    # The EQE is actually the IQE inside the fortran solver due to an error in the naming --> to be changed
    junction.eqe = interp1d(options.wavelength, output['QE']['EQE'], kind='linear', bounds_error=False,
                            assume_sorted=True, fill_value=(output['QE']['EQE'][0], output['QE']['EQE'][-1]))


# ----
# Functions for dumping data from the fortran variables
def DumpInputProperties():
    output = {}
    output['x'] = dd.get('x')[0:dd.m + 1]
    output['Xi'] = dd.get('xi')[0:dd.m + 1]
    output['Eg'] = dd.get('eg')[0:dd.m + 1]
    output['Nc'] = dd.get('nc')[0:dd.m + 1]
    output['Nv'] = dd.get('nv')[0:dd.m + 1]
    output['Nd'] = dd.get('nd')[0:dd.m + 1]
    output['Na'] = dd.get('na')[0:dd.m + 1]

    return output


def DumpBandStructure():
    output = {}
    output['x'] = dd.get('x')[0:dd.m + 1]
    output['n'] = dd.get('n')[0:dd.m + 1]
    output['p'] = dd.get('p')[0:dd.m + 1]
    output['ni'] = dd.get('ni')[0:dd.m + 1]
    output['Rho'] = dd.get('rho')[0:dd.m + 1]
    output['Efe'] = dd.get('efe')[0:dd.m + 1]
    output['Efh'] = dd.get('efh')[0:dd.m + 1]
    output['potential'] = dd.get('psi')[0:dd.m + 1]
    output['Ec'] = dd.get('ec')[0:dd.m + 1]
    output['Ev'] = dd.get('ev')[0:dd.m + 1]
    output['GR'] = dd.get('gr')[0:dd.m + 1]
    output['G'] = dd.get('g')[0:dd.m + 1]
    output['Rrad'] = dd.get('rrad')[0:dd.m + 1]
    output['Rsrh'] = dd.get('rsrh')[0:dd.m + 1]
    output['Raug'] = dd.get('raug')[0:dd.m + 1]

    return output


def DumpIV(IV_info=False):
    # Depending of having PN or NP the calculation of the MPP is a bit different. We move everithing to the 1st quadrant and then send it back to normal
    Nd = dd.get('nd')[0:dd.m + 1][0]
    Na = dd.get('na')[0:dd.m + 1][0]
    s = Nd > Na

    output = {}
    output['V'] = (-1) ** s * dd.get('volt')[1:dd.nvolt + 1]
    output['J'] = dd.get('jtot')[1:dd.nvolt + 1]
    output['Jrad'] = dd.get('jrad')[1:dd.nvolt + 1]
    output['Jsrh'] = dd.get('jsrh')[1:dd.nvolt + 1]
    output['Jaug'] = dd.get('jaug')[1:dd.nvolt + 1]
    output['Jsur'] = dd.get('jsur')[1:dd.nvolt + 1]

    if IV_info:
        # We calculate the solar cell parameters
        output['Jsc'] = -np.interp(0, output['V'], output['J'])  # dd.get('isc')[0]
        output['Voc'] = np.interp(0, output['J'][output['V'] > 0], output['V'][output['V'] > 0])  # dd.get('voc')[0]
        print(output['Voc'])

        maxPP = np.argmin(output['V'] * output['J'])
        Vmax = output['V'][maxPP - 3:maxPP + 3]
        Imax = output['J'][maxPP - 3:maxPP + 3]
        Pmax = Vmax * Imax

        poly = np.polyfit(Vmax, Pmax, 2)
        output['Vmpp'] = -poly[1] / (2 * poly[0])
        output['Jmpp'] = -output['J'][np.argmin(np.abs(output['V'] - output['Vmpp']))]
        output['FF'] = output['Vmpp'] * output['Jmpp'] / (output['Voc'] * output['Jsc'])

        print("Jsc  = %5.3f mA/cm2" % (output['Jsc'] / 10))
        print("Voc  = %4.3f  V" % (output['Voc']))
        print("FF   = %3.3f " % (output['FF']))
        print("Jmpp = %5.3f mA/cm2" % (output['Jmpp'] / 10))
        print("Vmpp = %4.3f  V" % (output['Vmpp']))
        print("Power= %5.3f mW/cm2" % (output['Jmpp'] * output['Vmpp'] / 10))

    # If NP, V and J should be in the 3rd quadrant
    output['V'] = (-1) ** s * output['V']
    output['J'] = (-1) ** s * output['J']
    output['Jrad'] = (-1) ** s * output['Jrad']
    output['Jsrh'] = (-1) ** s * output['Jsrh']
    output['Jaug'] = (-1) ** s * output['Jaug']
    output['Jsur'] = (-1) ** s * output['Jsur']

    return output


def DumpQE():
    output = {}
    numwl = dd.numwl + 1
    output['EQE'] = dd.get('iqe')[0:numwl]
    output['EQEsrh'] = dd.get('iqesrh')[0:numwl]
    output['EQErad'] = dd.get('iqerad')[0:numwl]
    output['EQEaug'] = dd.get('iqeaug')[0:numwl]
    output['EQEsurf'] = dd.get('iqesurf')[0:numwl]
    output['EQEsurb'] = dd.get('iqesurb')[0:numwl]

    return output


# ----
# Functions for setting the parameters controling the recombination, meshing and the numerial algorithm
def SetMeshParameters(options):
    dd.set('coarse', options.coarse)
    dd.set('fine', options.fine)
    dd.set('ultrafine', options.ultrafine)
    dd.set('growth', options.growth_rate)


def SetRecombinationParameters(options):
    dd.srh = options.srh
    dd.rad = options.rad
    dd.aug = options.aug
    dd.sur = options.sur
    dd.gen = options.gen


def SetConvergenceParameters(options):
    dd.nitermax = options.nitermax
    dd.set('clamp', options.clamp)
    dd.set('atol', options.ATol)
    dd.set('rtol', options.RTol)



import numpy as np
from solcore.solar_cell_solver import solar_cell_solver
from solcore.spice import solve_circuit


def solve_pv_module(solar_cell, options, totalcells=25, bias_start=0, bias_end=75, bias_step=0.1, jscSigma=2e-4,
                    shading=None):
    """ Calculate the IV curve of a PV module made of a certain number solar_cells connected in series in a single string. A certain dispersion to the distribution of photocurrents among cells and shadowing losses can be added to ahoeve more realistic results.

    :param solar_cell: A solar cell object containing all the junctions
    :param options: A state object with all the options needed to solve the solar cell IV
    :param totalcells: Total number of cells in a string
    :param bias_start: Initial voltage of the IV curve of the module
    :param bias_end: Final voltage of the IV curve of the module
    :param bias_step: Step for the bias
    :param jscSigma: Width of the distribution of short circuit currents around the ideal (eg: 0.02 = 2%). Default is set to 0.02% to aid convergence
    :param shading: Array containing the shading loses for each cell. It can be a 2D array (0 = no shading, 1 = full shading)
    :return: A tuple with the array of voltages, currents, a list with the Isc of each junction in each cell and the raw output from SPICE.
    """
    # We first start by the solar cell as if it were a normal, isolated cell
    solar_cell_solver(solar_cell, 'iv', user_options=options)

    # We don't care about this IV curve, in principle, but we care about some of the parameters calculated, like jsc,
    # j01 or j02 if calculated from detailed balance. We extract those parameters from the cell
    totaljuncs = solar_cell.junctions

    area = solar_cell.area if hasattr(solar_cell, 'area') else 1
    Rs = max(solar_cell.R_series / area, 1e-14)
    cell_temp = solar_cell.T - 273

    Isc_array = np.zeros(totaljuncs)
    I01_array = np.zeros(totaljuncs)
    n1_array = np.zeros(totaljuncs)
    I02_array = np.zeros(totaljuncs)
    n2_array = np.zeros(totaljuncs)
    Eg_array = np.zeros(totaljuncs)
    rsh_array = np.zeros(totaljuncs)

    for i in range(totaljuncs):

        try:
            Isc_array[i] = solar_cell(i).jsc * area
            I01_array[i] = solar_cell(i).j01 * area
            n1_array[i] = solar_cell(i).n1 if hasattr(solar_cell(i), 'n1') else 1
            I02_array[i] = solar_cell(i).j02 * area
            n2_array[i] = solar_cell(i).n2 if hasattr(solar_cell(i), 'n2') else 2
            Eg_array[i] = solar_cell(i).Eg
            rsh_array[i] = min(solar_cell(i).R_shunt / area, 1e14) if hasattr(solar_cell(i), 'R_shunt') else 1e14
        except AttributeError as err:
            raise AttributeError('ERROR in PV module: Junction is missing one essential argument. {}'.format(err))

    if shading is not None:
        shade = shading.flatten()
        assert len(shade) == totalcells
    else:
        shade = np.zeros(totalcells)

    # Calculate the probability distribution for the cells
    # Since all cells will not generate the same current, some random scatter can be added to the Isc.
    # note that the same random value is applied to all sub-cells in a particular junction.
    jsc_random_normal = np.random.normal(0.0, jscSigma, size=totalcells)

    # Shift the distribution so it is always <=1 so Isc never exceeds the ideal value,
    # and also take into account shading
    jsc_random_normal = abs(1 - max(jsc_random_normal) * jscSigma + jscSigma * jsc_random_normal) * (1 - shade)

    # We choose to keep track of the sub-cell Isc values so that they can be plotted at the end.
    all_Isc_values = []

    # Now we start to write the spice file
    spice_file = """ * SolCore spice Simulation \n.model bypassdiode d \n"""

    nodeCounter = 0
    junctionCounter = 0

    # We loop over all solar cells and all junctions within each solar cell
    for n in range(1, totalcells + 1):
        # Defines the lower connection for the bypass diode.
        lowerBypassConnection = nodeCounter
        temp_isc = []

        # Loop through junctions writing the spice code for each junction in turn
        for j in range(totaljuncs):
            junctionCounter += 1

            isc = Isc_array[j] * jsc_random_normal[n - 1]

            # Write the spice code for the junction to file
            spiceout = spice_junction(junctionCounter, nodeCounter, isc, I01_array[j], I02_array[j], n1_array[j],
                                      n2_array[j], Eg_array[j], rsh_array[j])

            spice_file = spice_file + spiceout
            nodeCounter = nodeCounter + 1
            temp_isc.append(isc)  # Save the isc value for later plotting

        # Entire cell has been built
        all_Isc_values.append(temp_isc)  # Add isc values to the main list.

        # Add the series resistance
        spiceout = 'r{0} {1} {2} {3}\n'.format(2 * junctionCounter - 1, nodeCounter+1, nodeCounter, Rs)
        nodeCounter = nodeCounter + 1
        spice_file = spice_file + spiceout

        # Connect bypass diode.  Connections are: lowerBypassConnection & nodeCounter
        spiceout = 'd{0} {1} {2} bypassdiode\n'.format(2 * junctionCounter + 1, lowerBypassConnection, nodeCounter)
        spice_file = spice_file + spiceout

        junctionCounter += 1

    spice_file = spice_file + ".OPTIONS TNOM=25 TEMP=" + str(cell_temp) + "\n"
    spice_file = spice_file + "vbias " + str(nodeCounter) + " 0\n"
    spice_file = spice_file + ".dc vbias " + str(bias_start) + " " + str(bias_end) + " " + str(bias_step) + "\n"
    spice_file = spice_file + ".plot dc i(vbias)\n"
    spice_file = spice_file + ".end \n\n"

    # And the definition of the spice file is finished. Now it's time to solve the problem
    raw_data = solve_circuit(spice_file)

    voltage = []
    current = []

    raw_data_in_lines = raw_data.split('\n')

    for line in raw_data_in_lines:
        if len(line) > 2 and line[1] in '0123456789':
            splitstring = line.split()
            voltage.append(float(splitstring[0]))
            current.append(float(splitstring[1]))

    return voltage, current, all_Isc_values, raw_data


def spice_junction(jc, nc, isc, j01, j02, n1, n2, Eg, rsh):
    """ Creates the string representation in SPICE of the junciton defined by the input values.

    :param jc: junction counter
    :param nc: node counter
    :param isc: Short circuit current
    :param j01: Reverse saturation current corresponding to ideality factor n1
    :param j02: Reverse saturation current corresponding to ideality factor n2
    :param n1: Ideality factor, typically = 1
    :param n2: Ideality factor, typically = 2
    :param Eg: Bandgap of the material the junction is made off
    :param rsh: Shunt resistance in the junction
    :return: A string representation of the junciton in SPICE
    """

    isource = 'i{0} {1} {2} dc {3}\n'.format(jc, nc, nc + 1, isc)
    d1 = 'd{0} {1} {2} diode{3} OFF\n'.format(2 * jc - 1, nc + 1, nc, 2 * jc - 1)
    d1deff = '.model diode{0} d(is={1},n={2},eg={3})\n'.format(2 * jc - 1, j01, n1, Eg)
    d2 = 'd{0} {1} {2} diode{3} OFF\n'.format(2 * jc, nc + 1, nc, 2 * jc)
    d2deff = '.model diode{0} d(is={1},n={2},eg={3})\n'.format(2 * jc, j02, n2, Eg)
    rshunt = 'r{0} {1} {2} {3}\n'.format(2 * jc, nc + 1, nc, rsh)

    junction = isource + d1 + d1deff + d2 + d2deff + rshunt


    return junction


if __name__ == '__main__':
    out = spice_junction(1, 1, 300, 1e-25, 1e-16, 1, 2, 1.4, 0.0001, 1e10)
    print(out)



import numpy as np
from numpy.linalg import eig
from functools import lru_cache

import solcore
from solcore.graphing import *
from solcore import constants
from solcore.crystals import traverse_brillouin, brillouin_critical_points, kvector
from solcore.science_tracker import science_reference

electron_charge = constants.q
hbar = constants.hbar
m0 = constants.electron_mass
q = constants.q


@lru_cache(maxsize=1000, typed=False)
def eight_band_strain_hamiltonian(kx, ky, kz, Ev0, Ec0, exx, ezz, me_eff, gamma1, gamma2, gamma3, a0, Delta, ac, av, b,
                                  Ep):
    """Hamiltonian to calculate cb, hh, lh, so bands and include strain for the biaxial (along [001]) special case.
    
    See Hamiltonian in ref. but remove row 1 and 6 and column 1 and 6.
    http://prb.aps.org/pdf/PRB/v73/i12/e125348"""
    science_reference("k.p Hamiltonian", "Stanko Tomic et al., Electronic structure of InyGa1yAs1xNxGaAs(N) quantum "
                                         "dots by ten-band k.p theory. Phys. Rev. B 73, 125348 (2006)")

    av = abs(
        av)  # The sign in Vurgaftmann is negative while Chuang (and Tomic) assumes it is possitve. We make sure it truly is.

    Eg = Ec0 - Ev0
    sqrt2 = np.sqrt(2)
    sqrt3 = np.sqrt(3)
    sqrt6 = np.sqrt(6)
    sqrt3o2 = np.sqrt(3 / 2)

    m0 = constants.electron_mass

    # The Kane matrix element (Vurgaftman 'interband matrix element')
    P0 = np.sqrt(Ep * constants.hbar ** 2 / (2 * m0))  # as a momentum

    # "Modified" Luttinger parameters
    gc = 1 / (me_eff / constants.electron_mass) - (Ep / 3) * (2 / Eg + 1 / (Eg + Delta))
    g1 = gamma1 - Ep / (
            3 * Eg + Delta)  # There was an error here. It was written as: g1 = gamma1 - Ep/(3*(Eg + Delta))
    g2 = gamma2 - Ep / (6 * Eg + 2 * Delta)
    g3 = gamma3 - Ep / (6 * Eg + 2 * Delta)

    Ck = constants.hbar ** 2 / (2 * m0)

    # Luttinger-Kohn terms
    Ok = lambda kx, ky, kz: Ck * gc * (kx ** 2 + ky ** 2 + kz ** 2)
    Pk = lambda kx, ky, kz: Ck * g1 * (kx ** 2 + ky ** 2 + kz ** 2)
    Qk = lambda kx, ky, kz: Ck * g2 * (kx ** 2 + ky ** 2 - 2 * kz ** 2)
    Rk = lambda kx, ky, kz: Ck * sqrt3 * (g2 * (kx ** 2 - ky ** 2) - 2 * 1j * g3 * kx * ky)
    Sk = lambda kx, ky, kz: Ck * sqrt6 * g3 * (kx - 1j * ky) * kz
    Tk = lambda kx, ky, kz: 1 / sqrt6 * P0 * (kx + 1j * ky)
    Uk = lambda kx, ky, kz: 1 / sqrt3 * P0 * kz

    # Conjugate
    Rkc = lambda kx, ky, kz: Ck * sqrt3 * (g2 * (kx ** 2 - ky ** 2) + 2 * 1j * g3 * kx * ky)
    Skc = lambda kx, ky, kz: Ck * sqrt6 * g3 * (kx + 1j * ky) * kz
    Tkc = lambda kx, ky, kz: 1 / sqrt6 * P0 * (kx - 1j * ky)

    # Strain terms (biaxial approximation makes many of these function zero)
    Oe = lambda exx, eyy, ezz: ac * (exx + eyy + ezz)
    Pe = lambda exx, eyy, ezz: -av * (exx + eyy + ezz)
    Qe = lambda exx, eyy, ezz: -b / 2 * (exx + eyy - 2 * ezz)
    # Re = lambda exx,eyy,ezz: 0
    # Se = lambda exx,eyy,ezz: 0
    # Tk = lambda exx,eyy,ezz: 0
    # Uk = lambda exx,eyy,ezz: 0

    # Complete terms
    O = lambda kx, ky, kz, exx, eyy, ezz: Ok(kx, ky, kz) + Oe(exx, eyy, ezz)
    P = lambda kx, ky, kz, exx, eyy, ezz: Pk(kx, ky, kz) + Pe(exx, eyy, ezz)
    Q = lambda kx, ky, kz, exx, eyy, ezz: Qk(kx, ky, kz) + Qe(exx, eyy, ezz)
    R = lambda kx, ky, kz, exx, eyy, ezz: Rk(kx, ky, kz)  # Here the strain term is zero
    S = lambda kx, ky, kz, exx, eyy, ezz: Sk(kx, ky, kz)  # Here the strain term is zero
    T = lambda kx, ky, kz, exx, eyy, ezz: Tk(kx, ky, kz)  # Here the strain term is zero
    U = lambda kx, ky, kz, exx, eyy, ezz: Uk(kx, ky, kz)  # Here the strain term is zero

    # Conjugates
    Rc = lambda kx, ky, kz, exx, eyy, ezz: Rkc(kx, ky, kz)  # Here the strain term is zero
    Sc = lambda kx, ky, kz, exx, eyy, ezz: Skc(kx, ky, kz)  # Here the strain term is zero
    Tc = lambda kx, ky, kz, exx, eyy, ezz: Tkc(kx, ky, kz)  # Here the strain term is zero

    # Bands
    Ecb = lambda kx, ky, kz, exx, eyy, ezz: Ec0 + O(kx, ky, kz, exx, eyy, ezz)
    Ehh = lambda kx, ky, kz, exx, eyy, ezz: Ev0 - (P(kx, ky, kz, exx, eyy, ezz) + Q(kx, ky, kz, exx, eyy, ezz))
    Elh = lambda kx, ky, kz, exx, eyy, ezz: Ev0 - (P(kx, ky, kz, exx, eyy, ezz) - Q(kx, ky, kz, exx, eyy, ezz))
    Eso = lambda kx, ky, kz, exx, eyy, ezz: Ev0 - (P(kx, ky, kz, exx, eyy, ezz) + Delta)

    eyy = exx
    p = P(kx, ky, kz, exx, eyy, ezz)
    q = Q(kx, ky, kz, exx, eyy, ezz)
    r = R(kx, ky, kz, exx, eyy, ezz)
    s = S(kx, ky, kz, exx, eyy, ezz)
    t = T(kx, ky, kz, exx, eyy, ezz)
    u = U(kx, ky, kz, exx, eyy, ezz)
    rc = Rc(kx, ky, kz, exx, eyy, ezz)
    sc = Sc(kx, ky, kz, exx, eyy, ezz)
    tc = Tc(kx, ky, kz, exx, eyy, ezz)
    cb = Ecb(kx, ky, kz, exx, eyy, ezz)
    hh = Ehh(kx, ky, kz, exx, eyy, ezz)
    lh = Elh(kx, ky, kz, exx, eyy, ezz)
    so = Eso(kx, ky, kz, exx, eyy, ezz)

    H_ST = np.mat([
        [cb, -sqrt3 * t, sqrt2 * u, -u, 0, 0, -tc, -sqrt2 * tc],
        [-sqrt3 * tc, hh, sqrt2 * s, -s, 0, 0, -r, -sqrt2 * r],
        [sqrt2 * u, sqrt2 * sc, lh, -sqrt2 * q, tc, r, 0, sqrt3 * s],
        [-u, -sc, -sqrt2 * q, so, sqrt2 * tc, sqrt2 * r, -sqrt3 * s, 0],
        [0, 0, t, sqrt2 * t, cb, -sqrt3 * tc, sqrt2 * u, -u],
        [0, 0, rc, sqrt2 * rc, -sqrt3 * t, hh, sqrt2 * sc, -sc],
        [-t, -rc, 0, -sqrt3 * sc, sqrt2 * u, sqrt2 * s, lh, -sqrt2 * q],
        [-sqrt2 * t, -sqrt2 * rc, sqrt3 * sc, 0, -u, -s, -sqrt2 * q, so]
    ])

    #     View hamiltonian, good for debugging
    #     if True:
    #         import pylab
    #         pylab.imshow(numpy.real(H_ST))
    #         pylab.show()

    H_ST = H_ST.transpose()
    E, Psi = eig(H_ST)  # do all the math

    bands = np.array((sorted(E.real)))
    return bands


def effective_mass_energy(k, m_eff, E0):
    return E0 + constants.hbar ** 2 / (2 * m_eff) * k ** 2


def kp_bands(material, host_material, kx=0, ky=0, kz=0, return_so=False, graph=False, fit_effective_mass=False,
             effective_mass_direction="X", additional_traverse=None):
    """

    :param material:
    :param host_material:
    :param kx:
    :param ky:
    :param kz:
    :param return_so:
    :param graph:
    :param fit_effective_mass:
    :param effective_mass_direction:
    :param additional_traverse:
    :return:
    """
    Delta = material.spin_orbit_splitting
    Eg = material.band_gap
    Ev0 = material.valence_band_offset
    Ec0 = material.valence_band_offset + Eg
    a0 = material.lattice_constant
    g1 = material.gamma1
    g2 = material.gamma2
    g3 = material.gamma3
    ac = material.a_c
    av = material.a_v
    b = material.b
    Ep = material.interband_matrix_element
    # print (ac, av, b, Ep, a0)
    me_eff = material.eff_mass_electron_Gamma * constants.electron_mass

    # Strain parameters
    exx = - (material.lattice_constant - host_material.lattice_constant) / material.lattice_constant
    ezz = - 2 * material.c12 / material.c11 * exx  # why this?
    # print("suspect stuff around here")
    zMax = np.pi / (a0) / 4
    kx = 0 * zMax
    ky = 0 * zMax

    result = eight_band_strain_hamiltonian(kx, ky, kz, Ev0, Ec0, exx, ezz, me_eff, g1, g2, g3, a0, Delta, ac, av, b, Ep)
    so, lh, hh, c = result[::2]
    if material.lattice_constant < host_material.lattice_constant:
        # print("CAUTION: BLINDLY SWAPPING HH,LH LABELS BASED ON LATTICE CONSTANT")
        lh, hh = hh, lh

    if fit_effective_mass:
        fitpoint_x = 0.05

        critical_points = brillouin_critical_points(material.lattice_constant)
        critical_gamma = critical_points["Gamma"]
        critical_L = critical_points[effective_mass_direction]

        fitpoint = critical_gamma + (critical_L - critical_gamma) * fitpoint_x
        k_cGx, k_cGy, k_cGz = critical_gamma
        k_fx, k_fy, k_fz = fitpoint
        bands_at_gamma = eight_band_strain_hamiltonian(k_cGx, k_cGy, k_cGz, Ev0, Ec0, exx, ezz, me_eff, g1, g2, g3, a0,
                                                       Delta, ac, av, b, Ep)
        bands_at_fitpoint = eight_band_strain_hamiltonian(k_fx, k_fy, k_fz, Ev0, Ec0, exx, ezz, me_eff, g1, g2, g3, a0,
                                                          Delta, ac, av, b, Ep)
        k_max = np.linalg.norm(fitpoint)

        effective_masses = []
        e0s = []
        for band in range(len(result)):
            fpE = bands_at_fitpoint[band]
            oE = bands_at_gamma[band]

            deltaE = fpE - oE
            m_eff_fit = constants.hbar ** 2 / (2 * deltaE) * k_max ** 2
            effective_masses.append(m_eff_fit)
            e0s.append(oE)

        m_eff_so, m_eff_lh, m_eff_hh, m_eff_c = effective_masses[::2]
        if material.lattice_constant < host_material.lattice_constant:
            # print("CAUTION: BLINDLY SWAPPING EFF MASS HH,LH LABELS BASED ON LATTICE CONSTANT")
            m_eff_lh, m_eff_hh = m_eff_hh, m_eff_lh

    if graph:
        traverse_order = ("Gamma", effective_mass_direction)
        if additional_traverse is not None:
            traverse_order = traverse_order + additional_traverse
        allk, x_graph, xticks = traverse_brillouin(a0, traverse_order=traverse_order, steps=300)  # ()

        ticks = [x[0] for x in xticks]
        labels = [x[1] for x in xticks]

        bands = []
        for kx, ky, kz in allk:
            bands.append(
                eight_band_strain_hamiltonian(kx, ky, kz, Ev0, Ec0, exx, ezz, me_eff, g1, g2, g3, a0, Delta, ac, av, b,
                                              Ep))

        g = []

        for i, b in enumerate(zip(*bands)):
            # print (asUnit(array(b),"eV"))
            if fit_effective_mass:
                g.append(GraphData(x_graph, effective_mass_energy(x_graph / fitpoint_x * k_max, effective_masses[i],
                                                                  e0s[i]) / electron_charge, color="red", linewidth=2,
                                   alpha=0.7))

            g.append(GraphData(x_graph, np.array(b) / electron_charge, color="black"))
        a = Graph(g, xticks=ticks, xticklabels=labels, ylabel="$E$ (eV)", xlabel="$k$", palette="hue wheel",
                  ylim=(-2, 1))
        # a = Graph(g, xticks = xticks, ylabel="$E$ (eV)", palette="hue wheel", ylim = (-2,1))
        # a = Graph(g,  ylabel="$E$ (eV)", palette="hue wheel", title="{} on {} -- Luttinger-Kohn w/Pikus-Bir(Stanko)".format(str(material), str(host_material)), ylim = (-2,1))
        a.draw()

    if fit_effective_mass:
        if return_so:
            return (c, hh, lh, so, np.abs(m_eff_c), np.abs(m_eff_hh), np.abs(m_eff_lh), np.abs(m_eff_so))
        else:
            return (c, hh, lh, np.abs(m_eff_c), np.abs(m_eff_hh), np.abs(m_eff_lh))

    if return_so:
        return (c, hh, lh, so)
    else:
        return (c, hh, lh)


def KPbands(material, host_material, return_edges_only=False, plot_result=False, t=0, p=np.pi / 2, fraction=0.2,
            points=50, vin=None):
    """ New version of the above function that produces either the band edges or the full bands in one specifc direction
    from Gamma.
    
    - Calculation of the full Brillouin zone is not allowed. After all, KP is only valid near k=0
    - The calculation of the effective masses is done externally by another function
    - The direction to calculate the bands is provided either with directory angles or a directory vector
    - The number of points of the bands and the fraction up to the border of the Brilluin zone in the given direction are also inputs
    - The default direction is X, as given by the angles theta (t) and phi (p)
    
    """

    Delta = material.spin_orbit_splitting
    Eg = material.band_gap
    Ev0 = material.valence_band_offset
    Ec0 = material.valence_band_offset + Eg
    a0 = material.lattice_constant
    g1 = material.gamma1
    g2 = material.gamma2
    g3 = material.gamma3
    ac = material.a_c
    av = material.a_v
    b = material.b
    Ep = material.interband_matrix_element
    me_eff = material.eff_mass_electron_Gamma * constants.electron_mass

    # Strain parameters
    exx = (host_material.lattice_constant - a0) / a0
    ezz = - 2 * material.c12 / material.c11 * exx

    if return_edges_only:
        # If we only want the band edges, we stop here
        result = eight_band_strain_hamiltonian(0, 0, 0, Ev0, Ec0, exx, ezz, me_eff, g1, g2, g3, a0, Delta, ac, av, b,
                                               Ep)
        so, lh, hh, c = result[::2]

        if a0 < host_material.lattice_constant:
            # We need to figure out a safer way of doing this
            # print("CAUTION: BLINDLY SWAPPING HH,LH LABELS BASED ON LATTICE CONSTANT")
            lh, hh = hh, lh

        return (c, hh, lh, so)

    allk = kvector(a=a0, t=t, p=p, fraction=fraction, points=points, vin=vin)

    bands = []
    for kx, ky, kz in allk:
        bands.append(
            eight_band_strain_hamiltonian(kx, ky, kz, Ev0, Ec0, exx, ezz, me_eff, g1, g2, g3, a0, Delta, ac, av, b, Ep))

    output = [np.array([np.linalg.norm(vect) for vect in allk])]
    for i, band in enumerate(zip(*bands)):
        output.append(np.array(band))

    if plot_result:
        import matplotlib.pyplot as plt
        for i in range(1, len(output)):
            plt.plot(output[0] * a0 / 2 / np.pi, output[i] / q, 'k')
        plt.xlabel('k (2*pi/a)')
        plt.ylabel('Energy (eV)')
        plt.show()

    output = np.array(output)
    return output


def fit_effective_masses(bands, material, host_material, plot_result=False, dk=0.026):
    from scipy.optimize import curve_fit

    masses = []
    a0 = material.lattice_constant

    # General maximum k value. The actual maximum value depends on the direction,
    # but the difference is generally not too big
    kmax = 2 * np.pi / a0

    # We give more weight in the fit to those k points near k=0
    weigth = np.exp(- bands[0] ** 2 / (dk * kmax) ** 2)
    for i in range(1, len(bands), 2):  # Bands are duplicated (at least for now) so we only fit half of them
        band = bands[i]

        # The initial guess for the effective mass (forward finite differences at k = 0)
        p0 = (bands[0][2] - bands[0][1]) * (bands[0][1] - bands[0][0]) / (
                    band[2] - 2 * band[1] + band[0]) * hbar ** 2 / m0

        # We consider only points with weight > 0.01 to speed up the algorithm and prevent convergence errors
        popt, pcov = curve_fit(parabolic, bands[0][weigth > 0.01], band[weigth > 0.01] - band[0], p0=p0,
                               sigma=weigth[weigth > 0.01])

        # Standard error in the calculation of the mass
        # perr = np.sqrt(np.diag(pcov))
        masses.append(popt[0])
        fit = band[0] + parabolic(bands[0], popt[0])

        if plot_result:
            import matplotlib.pyplot as plt
            plt.plot(bands[0] / kmax, band / q, 'k')
            plt.plot(bands[0] / kmax, fit / q, 'r')

    if a0 < host_material.lattice_constant:
        # We need to figure out a safer way of doing this
        # print("CAUTION: BLINDLY SWAPPING HH,LH LABELS BASED ON LATTICE CONSTANT")
        masses[1], masses[2] = masses[2], masses[1]

    if plot_result:
        plt.xlabel('k (2*pi/a)')
        plt.ylabel('Energy (eV)')
        plt.ylim(-2, 2)
        plt.show()

    masses = abs(np.array(masses)) * m0
    return tuple(masses[::-1].tolist())


def parabolic(k, mr):
    return hbar ** 2 / (2 * m0 * mr) * k ** 2


def average_inplane_effective_mass(material, host_material, averaging_points=10, dk=0.026):
    """ Calculates the average in-plane effective mass for the four bands, assuming always the XY plane. """

    # p = phi = Polar angle, always constant for the XY plane
    # t = theta = Azimuth. Given the symmetry of the problem, we need to average only 45˚ of the k space
    p = np.pi / 2
    t = np.linspace(0, np.pi / 4, averaging_points)

    mc = 0
    mhh = 0
    mlh = 0
    mso = 0

    # We calculate the effective mass on each direction and average them
    for tt in t:
        bands = KPbands(material, host_material, p=p, t=tt)
        result = fit_effective_masses(bands, material, host_material, dk=dk)

        mc = mc + result[0]
        mhh = mhh + result[1]
        mlh = mlh + result[2]
        mso = mso + result[3]

    mc = mc / averaging_points
    mhh = mhh / averaging_points
    mlh = mlh / averaging_points
    mso = mso / averaging_points

    return mc, mhh, mlh, mso


def kp8x8_bulk(material, host_material, averaging_points=10, dk=0.026):
    """ Combines the above functions to provide a compact result of the band edges and the effective masses. """

    mc, mhh, mlh, mso = average_inplane_effective_mass(material, host_material, averaging_points=averaging_points,
                                                       dk=dk)
    c, hh, lh, so = KPbands(material, host_material, return_edges_only=True)

    return c, hh, lh, so, mc, mhh, mlh, mso


if __name__ == "__main__":
    import solcore

    # Material parameters
    GaAs = solcore.material("GaAs")(T=300)
    InGaAs2 = solcore.material("InGaAs")(In=0.28, T=300)
    InGaAs3 = solcore.material("GaInAs")(In=0.2, T=300)
    GaAsP = solcore.material("GaInP")(In=0.7, T=300)

    bands = kp_bands(GaAs, InGaAs3, kx=0, ky=0, kz=0, graph=True, fit_effective_mass=True, effective_mass_direction="L",
                     return_so=True)

    # bands = KPbands(GaAs, GaAsP, fraction=0.2, plot_result=False)
    # result = fit_effective_masses(bands, GaAs, GaAsP, plot_result=False)
    #
    # print(result)



import copy
from solcore.graphing import Graph, GraphData, graph_defaults
from solcore.graphing.graph_support import open_with_os
from solcore.constants import *
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import FigureCanvasPdf
import tempfile

defaults = {
    "edit": lambda x, y: (x * 1e9, y / q),
    "xlabel": "Depth (nm)",
    "ylabel": "Energy (eV)",
}


def L(x, centre, hwhm):
    return 1 / pi * (0.5 * hwhm) / ((x - centre) ** 2 + (0.5 * hwhm) ** 2)  # Lorenzian (area normalised to 1)


def structure_graph(structure, **kwargs):
    global defaults
    options = copy.copy(defaults)
    options.update(kwargs)

    current_x = 0
    x = []
    conduction = []
    valence = []
    for layer in structure:
        x.append(current_x)
        conduction.append(layer.Ec)
        conduction.append(layer.Ec)
        valence.append(layer.Ev)
        valence.append(layer.Ev)
        current_x += layer.width
        x.append(current_x)

    # x = np.array(x)
    # conduction = np.array(conduction)
    # valence = np.array(valence)
    g = Graph(
        GraphData(x, conduction, label="Conduction Band", linewidth=2, color="black"),
        GraphData(x, valence, label="Valence Band", linewidth=2, color="black"),
        **options
    )
    return g


def normalise_psi(p):
    return p / (max(p) - min(p)) * q


def prepare_wavefunction_data(x, E, Psi, trim_levels_beyond, linewidth, color, scale, suppress_invert=False,
                              square=False, alpha=1, label="untitled"):
    data = []
    for e, psi in zip(E, Psi):
        norm_psi = normalise_psi(psi if not square else psi ** 2)
        trim_to = norm_psi ** 2 / q ** 2 > trim_levels_beyond ** 2
        trimmed_x = x[trim_to]
        norm_psi = norm_psi[trim_to]
        invert = -1 if norm_psi[0] < 0 and not suppress_invert and not square else 1

        linekwargs = {"linewidth": linewidth, "alpha": alpha, "label": label}
        data.append(GraphData(trimmed_x, e + norm_psi * scale * invert, color=color, **linekwargs))
    return data


def wavefunctions_graph(
        x, Ee=None,
        psi_e=None,
        Ehh=None,
        psi_hh=None,
        Elh=None,
        psi_lh=None,
        trim_levels_beyond=1e-3,
        linewidth=1,
        scale=0.06,
        colour_by_band=True,
        **kwargs):
    global defaults
    options = copy.copy(defaults)
    options.update(kwargs)

    data = []

    # normalise_psi = lambda p:p/numpy.sqrt(numpy.trapz(x=x*1e9,y=p**2))*q

    if Ee is not None: data.extend(
        prepare_wavefunction_data(x, Ee, psi_e, trim_levels_beyond, linewidth, "blue", scale, label="e"))
    if Ehh is not None: data.extend(
        prepare_wavefunction_data(x, Ehh, psi_hh, trim_levels_beyond, linewidth, "green", scale, label="hh"))
    if Elh is not None: data.extend(
        prepare_wavefunction_data(x, Elh, psi_lh, trim_levels_beyond, linewidth, "red", scale, label="lh"))

    g = Graph(data, **options)
    return g


def potentials_graph(x, Ve=None, Vhh=None, Vlh=None, color="grey", **kwargs):
    global defaults
    options = copy.copy(defaults)
    options.update(kwargs)
    data = []

    normalise_psi = lambda p: p * q * 5e-6

    if Ve is not None: data.append(GraphData(x, Ve, color=color, linewidth=2, label="Ve"))
    if Vlh is not None: data.append(GraphData(x, Vlh, color=color, linewidth=2, dashes=[1, 1], label="Vlh"))
    if Vhh is not None: data.append(GraphData(x, Vhh, color=color, linewidth=2, label="Vhh"))

    g = Graph(data, **options)
    return g


def schrodinger_graph(schrodinger_result, **kwargs):
    potentials_kwargs = copy.copy(schrodinger_result["potentials"])
    potentials_kwargs.update(kwargs)
    potential_graph = potentials_graph(**potentials_kwargs)

    wavefunctions_kwargs = copy.copy(schrodinger_result["wavefunctions"])
    wavefunctions_kwargs.update(kwargs)
    wavefunctions_kwargs.update(schrodinger_result["E"])

    w = wavefunctions_graph(figure=potential_graph.figure, trim_levels_beyond=1e-3, **wavefunctions_kwargs)
    return w


def split_schrodinger_graph(schrodinger_result,
                            trim_levels_beyond=1e-2,
                            linewidth=1,
                            scale=0.03,
                            suppress_invert=False,
                            probability_density=False,
                            wfalpha=0.8,
                            potentialalpha=0.8,

                            **kwargs):
    options = copy.copy(defaults)
    options["square"] = False

    defaults.update(kwargs)
    potentials = schrodinger_result["potentials"]
    wavefunctions = schrodinger_result["wavefunctions"]
    energy_levels = schrodinger_result["E"]
    x = schrodinger_result["x"]

    # This is used when doing the 4x4 kp calculation rather than the single band calculation
    if 'EU' in energy_levels.keys():
        energy_levels['Ehh'] = energy_levels['EU']
        energy_levels['Elh'] = energy_levels['EU']
        wavefunctions["psi_hh"] = wavefunctions["psi_g1"]
        wavefunctions["psi_lh"] = wavefunctions["psi_g2"]

    conduction_data = [GraphData(x, potentials["Ve"], linewidth=2, color="grey", alpha=potentialalpha)]
    valence_data = [
        GraphData(x, potentials["Vlh"], linewidth=2, color="grey", alpha=potentialalpha, dashes=[1, 1], label="Vlh"),
        GraphData(x, potentials["Vhh"], linewidth=2, color="grey", alpha=potentialalpha, label="Vhh")
    ]
    # normalise_psi = lambda p:p/numpy.sqrt(numpy.trapz(x=x*1e9,y=p**2))*q

    conduction_data.extend(
        prepare_wavefunction_data(x, energy_levels["Ee"], wavefunctions["psi_e"], trim_levels_beyond, linewidth, "blue",
                                  scale, suppress_invert, alpha=wfalpha, square=probability_density))
    valence_data.extend(
        prepare_wavefunction_data(x, energy_levels["Ehh"], wavefunctions["psi_hh"], trim_levels_beyond, linewidth,
                                  "green",
                                  scale, suppress_invert, alpha=wfalpha, square=probability_density))
    valence_data.extend(
        prepare_wavefunction_data(x, energy_levels["Elh"], wavefunctions["psi_lh"], trim_levels_beyond, linewidth,
                                  "red",
                                  scale, suppress_invert, alpha=wfalpha, square=probability_density))

    g = Graph(valence_data, subplot=212, **options)
    del options["xlabel"]
    g.add_subplot(conduction_data, subplot=211, **options)
    # g.axis.get_xaxis().set_ticklabels([])
    return g


def split_schrodinger_graph_potentials(schrodinger_result,
                                       trim_levels_beyond=1e-2,
                                       linewidth=1,
                                       scale=0.3,
                                       suppress_invert=False,
                                       probability_density=False,
                                       wfalpha=0.8,
                                       potentialalpha=0.8,
                                       **kwargs):
    defaults = {'step': 0.002, 'margin': 0.02, 'pdf': False, 'show': False, 'dpi': 100, 'fontsize': 12,
                'figsize': (7, 6)}
    options = copy.copy(defaults)
    options["square"] = False

    defaults.update(kwargs)
    potentials = schrodinger_result["potentials"]
    wavefunctions = schrodinger_result["wavefunctions"]
    energy_levels = schrodinger_result["E"]
    x = schrodinger_result["x"]

    # This is used when doing the 4x4 kp calculation rather than the single band calculation
    if 'EU' in energy_levels.keys():
        energy_levels['Ehh'] = energy_levels['EU']
        energy_levels['Elh'] = energy_levels['EU']
        wavefunctions["psi_hh"] = wavefunctions["psi_g1"]
        wavefunctions["psi_lh"] = wavefunctions["psi_g2"]

    fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, figsize=defaults['figsize'], dpi=defaults['dpi'])

    ax1.plot(x * 1e9, potentials["Ve"] / q, 'k', linewidth=2, label='Ve')
    ax1.set_ylabel('Energy (eV)', fontsize=defaults["fontsize"])
    ax1.tick_params(labelsize=defaults["fontsize"])
    ax1.grid(color='grey', linestyle='--', linewidth=0.5)

    ax2.plot(x * 1e9, potentials["Vlh"] / q, 'k--', linewidth=2, label="Vlh"),
    ax2.plot(x * 1e9, potentials["Vhh"] / q, 'k', linewidth=2, label="Vhh")
    ax2.set_ylabel('Energy (eV)', fontsize=defaults["fontsize"])
    ax2.set_xlabel('Possition (nm)', fontsize=defaults["fontsize"])
    ax2.tick_params(labelsize=defaults["fontsize"])
    ax2.grid(color='grey', linestyle='--', linewidth=0.5)

    e_data = prepare_wavefunction_data_only(x, energy_levels["Ee"] / q, wavefunctions["psi_e"], trim_levels_beyond,
                                            linewidth, "blue",
                                            0.03 / q, suppress_invert, alpha=wfalpha, square=probability_density)
    hh_data = prepare_wavefunction_data_only(x, energy_levels["Ehh"] / q, wavefunctions["psi_hh"], trim_levels_beyond,
                                             linewidth,
                                             "green", 0.03 / q, suppress_invert, alpha=wfalpha,
                                             square=probability_density)
    lh_data = prepare_wavefunction_data_only(x, energy_levels["Elh"] / q, wavefunctions["psi_lh"], trim_levels_beyond,
                                             linewidth,
                                             "red", 0.03 / q, suppress_invert, alpha=wfalpha,
                                             square=probability_density)

    for x, y in e_data:
        ax1.plot(x * 1e9, y, 'blue', linewidth=2, label='e')
    for x, y in hh_data:
        ax2.plot(x * 1e9, y, 'green', linewidth=2, label='hh')
    for x, y in lh_data:
        ax2.plot(x * 1e9, y, 'red', linewidth=2, label='lh')

    if defaults["show"]:
        plt.tight_layout()
        plt.show()

    if defaults["pdf"]:
        handle, path = tempfile.mkstemp(prefix="tmp_solcore_", suffix=".%s" % graph_defaults["format"])
        canvas = FigureCanvasPdf(fig)
        canvas.print_figure(path, dpi=defaults["dpi"], bbox_inches='tight')
        open_with_os(path)


def split_schrodinger_graph_LDOS(schrodinger_result, **kwargs):
    defaults = {'step': 0.001, 'margin': 0.02, 'pdf': False, 'show': False, 'dpi': 100, 'fontsize': 12,
                'figsize': (7, 6)}
    defaults.update(kwargs)

    effective_masses = schrodinger_result["effective_masses"]
    potentials = schrodinger_result["potentials"]
    wavefunctions = schrodinger_result["wavefunctions"]
    energy_levels = schrodinger_result["E"]
    x = schrodinger_result["x"]

    if 'EU' in energy_levels.keys():
        energy_levels['Ehh'] = energy_levels['EU']
        energy_levels['Elh'] = energy_levels['EU']
        wavefunctions["psi_hh"] = wavefunctions["psi_g1"]
        wavefunctions["psi_lh"] = wavefunctions["psi_g2"]

    Ee, LDOSe = LDOS1D_e(x, energy_levels, wavefunctions, effective_masses, defaults['step'], defaults['margin'])
    Eh, LDOSh = LDOS1D_h(x, energy_levels, wavefunctions, effective_masses, defaults['step'], defaults['margin'])

    fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, figsize=defaults['figsize'], dpi=defaults['dpi'])

    ax1.contourf(x * 1e9, Ee / q, LDOSe, 100, cmap='gnuplot2_r', vmin=0, vmax=max(LDOSe.flatten()) * 1.2)
    ax1.plot(x * 1e9, potentials["Ve"] / q, 'k', linewidth=2, label='Ve')
    ax1.set_ylabel('Energy (eV)', fontsize=defaults["fontsize"])
    ax1.tick_params(labelsize=defaults["fontsize"])

    ax2.contourf(x * 1e9, Eh / q, LDOSh, 100, cmap='gnuplot2_r', vmin=0, vmax=max(LDOSh.flatten()) * 1.2)
    ax2.plot(x * 1e9, potentials["Vlh"] / q, 'k--', linewidth=2, label="Vlh"),
    ax2.plot(x * 1e9, potentials["Vhh"] / q, 'k', linewidth=2, label="Vhh")
    ax2.set_ylabel('Energy (eV)', fontsize=defaults["fontsize"])
    ax2.set_xlabel('Possition (nm)', fontsize=defaults["fontsize"])
    ax2.tick_params(labelsize=defaults["fontsize"])

    if defaults["show"]:
        plt.tight_layout()
        plt.show()

    if defaults["pdf"]:
        handle, path = tempfile.mkstemp(prefix="tmp_solcore_", suffix=".%s" % graph_defaults["format"])
        canvas = FigureCanvasPdf(fig)
        canvas.print_figure(path, dpi=defaults["dpi"], bbox_inches='tight')
        open_with_os(path)

    return Ee, LDOSe, Eh, LDOSh


def LDOS1D_e(x, E, psi, m, step=0.001, margin=0.02, broad=0.005):
    Emax = max(E['Ee']) + margin * q
    Emin = min(E['Ee']) - margin * q

    energy = np.arange(Emin, Emax, step * q)
    LDOS = np.zeros((len(energy), len(x)))

    for i, ee in enumerate(E['Ee']):
        m_plane = calculate_in_plane_masses(x, psi['psi_e'], m['me'])
        LDOS = LDOS + m_plane[i] / pi / hbar ** 2 * np.outer(L(energy, ee, broad * q), psi['psi_e'][i] ** 2)

    return energy, LDOS


def LDOS1D_h(x, E, psi, m, step=0.001, margin=0.02, broad=0.005):
    Emax = max(max(E['Ehh']), max(E['Elh'])) + margin * q
    Emin = min(min(E['Ehh']), min(E['Elh'])) - margin * q

    energy = np.arange(Emin, Emax, step * q)
    LDOS = np.zeros((len(energy), len(x)))

    for i, ee in enumerate(E['Ehh']):
        m_plane = calculate_in_plane_masses(x, psi['psi_hh'], m['mhh'])
        LDOS = LDOS + m_plane[i] / pi / hbar ** 2 * np.outer(L(energy, ee, broad * q), psi['psi_hh'][i] ** 2)

    for i, ee in enumerate(E['Elh']):
        m_plane = calculate_in_plane_masses(x, psi['psi_lh'], m['mlh'])
        LDOS = LDOS + m_plane[i] / pi / hbar ** 2 * np.outer(L(energy, ee, broad * q), psi['psi_lh'][i] ** 2)

    return energy, LDOS


def calculate_in_plane_masses(x, psi, m):
    """ Calculates the in-plane effective mass for each level, considering that the wavefunction leaks into the barriers."""

    m_out = []
    for ps in psi:
        m_out.append(np.trapz(ps ** 2 * m, x))

    return m_out


def prepare_wavefunction_data_only(x, E, Psi, trim_levels_beyond, linewidth, color, scale, suppress_invert=False,
                                   square=False, alpha=1, label="untitled"):
    data = []
    for e, psi in zip(E, Psi):
        norm_psi = normalise_psi(psi if not square else psi ** 2)
        trim_to = norm_psi ** 2 / q ** 2 > trim_levels_beyond ** 2
        trimmed_x = x[trim_to]
        norm_psi = norm_psi[trim_to]
        invert = -1 if norm_psi[0] < 0 and not suppress_invert and not square else 1

        data.append((trimmed_x, e + norm_psi * scale * invert))

    return data


"""


def LDOS_h(x, E, psi, m, step=0.002, margin=0.05):
    Emax = max(max(E['Ehh']), max(E['Elh'])) + margin * q
    Emin = min(min(E['Ehh']), min(E['Elh'])) - margin * q

    energy = np.arange(Emin, Emax, step * q)
    LDOS = np.zeros((len(energy), len(x)))

    for i, ee in enumerate(E['Ehh']):
        m_plane = calculate_in_plane_masses(x, psi['psi_hh'], m['mhh'])
        LDOS = LDOS + m_plane[i] / pi / hbar ** 2 * np.outer((energy <= ee), psi['psi_hh'][i] ** 2)

    for i, ee in enumerate(E['Elh']):
        m_plane = calculate_in_plane_masses(x, psi['psi_lh'], m['mlh'])
        LDOS = LDOS + m_plane[i] / pi / hbar ** 2 * np.outer((energy <= ee), psi['psi_lh'][i] ** 2)

    return energy, LDOS

def LDOS_e(x, E, psi, m, step=0.002, margin=0.05):
    Emax = max(E['Ee']) + margin * q
    Emin = min(E['Ee']) - margin * q

    energy = np.arange(Emin, Emax, step * q)
    LDOS = np.zeros((len(energy), len(x)))

    for i, ee in enumerate(E['Ee']):
        m_plane = calculate_in_plane_masses(x, psi['psi_e'], m['me'])
        LDOS = LDOS + m_plane[i] / pi / hbar ** 2 * np.outer((energy >= ee), psi['psi_e'][i] ** 2)

    # plt.contourf(x, energy/q, DOS, 50, cmap='gnuplot2_r', vmin=0, vmax=max(DOS.flatten())*1.2)
    # plt.clim(0, max(DOS.flatten())*1.2)
    # plt.show()

    return energy, LDOS
    


"""



from .IV import iv_multijunction
from .QE import spectral_response_all_junctions
from .detailed_balance import iv_detailed_balance, qe_detailed_balance, absorptance_detailed_balance, db_options
from .depletion_approximation import iv_depletion, qe_depletion
from .diode_equation import iv_2diode
from .tunnel_junctions import resistive_tunnel_junction, parametric_tunnel_junction, external_tunnel_junction


from copy import copy
import numpy

from matplotlib.patches import Rectangle
from matplotlib import font_manager

line_defaults = {
    "linewidth": 0.5,
}


class GraphData_Base(object):
    def __init__(self, *data, color_index=None, edit=None, **options):
        if not (hasattr(self, "suppress_defaults") and self.suppress_defaults):
            self.line_format = copy(line_defaults)
        else:
            self.line_format = {}

        self.color_index = color_index
        if len(data) == 1:
            self.data = numpy.array(data[0])
        else:
            self.data = numpy.array(data)
        self.line_format.update(options)
        if edit is not None:
            self.edit = edit

    def draw(self, axis, figure=None):
        raise NotImplementedError("Subclasses must define the draw method")

    @property
    def data(self):
        if hasattr(self, "edit"):
            if len(self._data) == 2:
                return self.edit(*self._data)
            if len(self._data) == 3:
                x, y1 = self.edit(self._data[0], self._data[1])
                x, y2 = self.edit(self._data[0], self._data[2])
                return numpy.array((x, y1, y2))
        return self._data

    @data.setter
    def data(self, new_value):
        self._data = new_value
        # def legend_entry(self):
        #     return self.line_format["legend"]


class GraphDataHRect(GraphData_Base):
    def draw(self, axis, figure=None):
        x1, x2 = self.data
        axis.axhspan(x1, x2, **self.line_format)


class GraphDataVRect(GraphData_Base):
    def draw(self, axis, figure=None):
        x1, x2 = self.data
        axis.axvspan(x1, x2, **self.line_format)
        # if "label" in self.line_format:
        #     patch_color = {}
        #     patch_color.update(self.line_format)
        #     p = Rectangle((0, 0), 1, 1, **patch_color)
        #     return p, self.line_format["label"]


class GraphData(GraphData_Base):
    def draw(self, axis, figure=None):
        x, y = self.data
        axis.plot(x, y, **self.line_format)


class GraphDataErrorbars(GraphData_Base):
    def draw(self, axis, figure=None):
        assert (
        "xerr" in self.line_format or "yerr" in self.line_format), "GraphDataErrorbars requres xerr, yerr or both to be set"
        x, y = self.data
        axis.errorbar(x, y, **self.line_format)


class GraphDataBetween(GraphData_Base):
    def draw(self, axis, figure=None):
        x, y1, y2 = self.data
        axis.fill_between(x, y1, y2, **self.line_format)
        if "label" in self.line_format:
            patch_color = {}
            patch_color.update(self.line_format)
            p = Rectangle((0, 0), 1, 1, **patch_color)
            return p, self.line_format["label"]


class LegendEntry(GraphData_Base):
    def draw(self, axis, figure=None):
        p = Rectangle((0, 0), 1, 1, edgecolor='none', facecolor='none')

        return p, self.data


class GraphDataLabels(GraphData_Base):
    def __init__(self, *data, color_index=None, font_size=None, fontname=None, weight=None, font_color=None, **options):
        font_args = {}

        if font_size is not None:
            font_args["size"] = font_size
        if fontname is not None:
            font_args["font_name"] = fontname
        if weight is not None:
            font_args["weight"] = weight
        if font_color is not None:
            font_args["color"] = font_color

        self.font_properties = font_manager.FontProperties(**font_args)
        GraphData_Base.__init__(self, *data, color_index=color_index, **options)

    def draw(self, axis, figure=None):
        x, y, labels = self.data
        axis.scatter(x, y, **self.line_format)
        for X, Y, L in zip(x, y, labels):
            axis.annotate(" " + L, xy=(X, Y), fontproperties=self.font_properties)
            # if "label" in self.line_format:
            #     patch_color = {}
            #     patch_color.update(self.line_format)
            #     p = Rectangle((0, 0), 1, 1, **patch_color)
            #     return p, self.line_format["label"]


class Text(GraphData_Base):
    def draw(self, axis, figure=None, ):
        print(self._data[0], self._data[1], self._data[2])
        figure.text(self._data[0], self._data[1], self._data[2])


class GraphDataImage(GraphData_Base):
    suppress_defaults = True

    def draw(self, axis, figure=None):
        z = self.data
        image = axis.imshow(z, **{k: self.line_format[k] for k in self.line_format.keys() if k not in ["color", ]})



import os
from datetime import datetime

SOLCORE_ROOT = os.path.split(__file__)[0]
default_config = os.path.join(SOLCORE_ROOT, "solcore_config.txt")
user_path = os.environ.get("SOLCORE_USER_DATA")
if user_path is None:
    user_path = os.path.join(os.path.expanduser("~"), ".solcore")
if not os.path.exists(user_path):
    os.mkdir(user_path)
user_config = os.path.join(user_path, "solcore_config.txt")

from .config_tools import SolcoreConfig

config = SolcoreConfig(default_config, user_config)
verbose = config.verbose_loading()
__version__ = config.version()

if config.welcome_message():
    print(
        f"""\n\tWelcome to Solcore - version {config.version()}
\tCopyright (c) {datetime.today().year}, Imperial College London. All rights reserved.
\tSoftware released under the GNU Lesser General Public License.\n"""
    )

from solcore.units_system import UnitsSystem

# First we populate the Units system:
us = UnitsSystem(config.units)
config.register_observer("Units", us.read)

# And now we load some functions form it.
si = us.si
asUnit = us.asUnit
siUnits = us.siUnits
sensibleUnits = us.sensibleUnits
siUnitFromString = us.siUnitFromString
convert = us.convert
guess_dimension = us.guess_dimension
nmJ = us.nmJ
mJ = us.mJ
eVnm = us.eVnm
nmHz = us.nmHz
spectral_conversion_nm_ev = us.spectral_conversion_nm_ev
spectral_conversion_nm_hz = us.spectral_conversion_nm_hz
eV = us.eV

# And the same with the Parameter system
from solcore.parameter_system import ParameterSystem

ps = ParameterSystem(config.parameters)
config.register_observer("Parameters", ps.read)
get_parameter = ps.get_parameter

# And the same with the Materials system
from solcore.material_system import MaterialSystem

ms = MaterialSystem(config.materials)
config.register_observer("Materials", ms.read)
material = ms.material



import numpy as np
import solcore
from solcore.state import State

"""
A comment on strain calculations. -- DJF July 2012

It is really easy to make mistakes when applying the strain code because the 
sign of the parameters is crucial and there appears to be no standard sign 
convention in the literature.

The procedure to align the heterointerfaces:

0) Align unstrained heterosructure using valence band offsets.

1) Shift conduction band using 'a_c' deformation potential (compressive/tensile = up/down)

2) Shift HH and LH bands using 'a_v' deformation potential (compressive/tensile = down/up)

3) Split the valence bands using the 'b' deformation potential (compressive/tensile = (HH)up, (LH)down/(HH)down, (LH)up)

"""

def strain_calculation_asserts(k, should_print=False):
    """Will check that the sign of parameters for the strain calcuations is correct."""
    if should_print: print(k)
    assert_sign = dict()

    # Sign for compressive strain
    assert_sign['epsilon'] = -1
    assert_sign['e_xx'] = -1
    assert_sign['e_yy'] = -1
    assert_sign['e_zz'] = 1
    assert_sign['Pe'] = 1
    assert_sign['Qe'] = -1
    assert_sign['delta_Ec'] = 1
    # assert_sign['deltaEhh'] = -1 # Sign of delta_Ehh not guaranteed
    # assert_sign['deltaElh'] = -1 # Sign of delta_Elh not guaranteed

    # Filp sign of these parameters for tensile strain
    if k.epsilon > 0:
        for key in assert_sign.keys():
            assert_sign[key] = assert_sign[key] * -1

    # Sign constant regardless of compressive or tensile strain
    assert_sign['ac'] = -1
    assert_sign['av'] = 1
    assert_sign['b'] = -1

    if should_print:
        print()
        print("Checking sign of strain parameters:")
    for key in assert_sign.keys():

        # If the material of the substrate form part of the layers, the sign makes no sense. We skip that.
        if k[key] == 0:
            continue

        if np.sign(k[key]) == assert_sign[key]:
            if should_print:
                print(key, "\tOK")
        else:
            if should_print:
                print(key, "\tPROBLEM")
            raise ValueError("The parameter `%s' used by the qwell module appears to have the wrong sign." % key)


def strain_calculation_parameters(substrate_material, layer_material, should_print=False, SO=False):
    """Will extract materials parameters and perform a bit of conditioning to the values.
    
    Returns a solcore State object with the following keys:
    
        -- ac, the conduction band deformation potential
        -- av, the valence band deformation potential
        -- b,  the valence band splitting deformation potential
        -- C11, element of the elastic stiffness tensor
        -- C12, element of the elastic stiffness tensor
        -- a0, the unstrained substrate lattice constant
        -- a, the unstrained layer lattice constant
        -- epsilon, in-plain strain
        -- epsilon_perp, out-plain strain
        -- e_xx, in-plain strain (e_xx = epsilon)
        -- e_yy, in-plain strain (e_yy = epsilon)
        -- e_zz, out-plain strain (e_zz = epsilon_perp)
        -- Tre, element of come matrix (Tre = e_xx + e_yy + e_zz)
        -- Pe, parameter use by Chuang
        -- Qe, parameter use by Chuang
        -- cb_hydrostatic_shift, CB moved by this amount
        -- vb_hydrostatic_shift, VB moved by this amount
        -- vb_shear_splitting, VB split by this amount (i.e. HH/LH separation)
        -- delta_Ec, final conduction band shift
        -- delta_Elh, final light hole band shift
        -- delta_Ehh, final heavy hole band shift
    
    Care has to be taken when calculating the energy shifts because different 
    sign conversion are used by different authors. Here we use the approach of
    S. L. Chuang, 'Physics of optoelectronic devices'. 
    
    Note that this requires that the 'a_v' deformation potential to be 
    positive, where as Vurgaftman defines this a negative!
    """

    sub = substrate_material
    mat = layer_material
    k = State()

    # deformation potentials
    k.av = abs(mat.a_v)  # make sure that av is positive for this calculation
    k.ac = mat.a_c  # Vurgaftman uses the convention that this is negative
    k.b = mat.b

    # Matrix elements from elastic stiffness tensor
    k.C11 = mat.c11
    k.C12 = mat.c12
    if should_print: print(sub, mat)
    # Strain fractions
    k.a0 = sub.lattice_constant
    k.a = mat.lattice_constant
    k.epsilon = (k.a0 - k.a) / k.a  # in-plain strain
    k.epsilon_perp = -2 * k.C12 / k.C11 * k.epsilon  # out-plain
    k.e_xx = k.epsilon
    k.e_yy = k.epsilon
    k.e_zz = k.epsilon_perp
    k.Tre = (k.e_xx + k.e_yy + k.e_zz)
    k.Pe = -k.av * k.Tre
    k.Qe = -k.b / 2 * (k.e_xx + k.e_yy - 2 * k.e_zz)

    k.cb_hydrostatic_shift = k.ac * k.Tre
    k.vb_hydrostatic_shift = k.av * k.Tre
    k.vb_shear_splitting = 2 * k.b * (1 + 2 * k.C12 / k.C11) * k.epsilon

    # Shifts and splittings
    k.delta_Ec = k.ac * k.Tre

    if should_print: print(k.ac, k.Tre)

    k.delta_Ehh = -k.Pe - k.Qe
    k.delta_Elh = -k.Pe + k.Qe
    k.delta_Eso = 0.0

    if SO:
        k.delta = mat.spin_orbit_splitting
        shift = k.delta ** 2 + 2 * k.delta * k.Qe + 9 * k.Qe ** 2
        k.delta_Elh = -k.Pe + 0.5 * (k.Qe - k.delta + np.sqrt(shift))
        k.delta_Eso = -k.Pe + 0.5 * (k.Qe - k.delta - np.sqrt(shift))

    strain_calculation_asserts(k, should_print=should_print)

    if should_print:
        print()
        print("Lattice:")
        print("a0", k.a0)
        print("a", k.a)
        print()
        print("Deformation potentials:")
        print("ac = ", solcore.asUnit(k.ac, 'eV'))
        print("av = ", solcore.asUnit(k.av, 'eV'))
        print("ac - av = ", solcore.asUnit(k.ac - k.av, 'eV'))
        print("b = ", solcore.asUnit(k.b, 'eV'))
        print()
        print("Matrix elements from elastic stiffness tensor:")
        print("C_11 = ", solcore.asUnit(k.C11, "GPa"))
        print("C_12 = ", solcore.asUnit(k.C12, "GPa"))
        print()
        print("Strain fractions:")
        print("e_xx = e_yy = epsilon = ", k.epsilon)
        print("e_zz = epsilon_perp = ", k.epsilon_perp)
        print("e_xx + e_yy + e_zz = Tre = ", k.Tre)
        print()
        print("Shifts and splittings:")
        print("Pe = -av * Tre = ", solcore.asUnit(k.Pe, 'eV'))
        print("Qe = -b/2*(e_xx + e_yy - 2*e_zz) = ", solcore.asUnit(k.Qe, 'eV'))
        print("dEc = ac * Tre = ", solcore.asUnit(k.delta_Ec, 'eV'))
        print("dEhh = av * Tre + b[1 + 2*C_11/C_12]*epsilon = -Pe - Qe = ", solcore.asUnit(k.delta_Ehh, 'eV'))
        print("dElh = av * Tre - b[1 + 2*C_11/C_12]*epsilon = -Pe + Qe = ", solcore.asUnit(k.delta_Elh, 'eV'))
        print()

    return k



""" Solves an electrical circuit using SPICE. The input must be a correctly formated SPICE file and the output will be
the raw result from SPICE or after some processing, depending of what is requested. """

import numpy
import os
import subprocess
import tempfile
import solcore


class SpiceSolverError(Exception):
    pass


class spice:
    engine = os.environ.get("SOLCORE_SPICE") if os.environ.get("SOLCORE_SPICE") is not None else solcore.config.spice()
    input_file = "current_spice.cir"
    output_file = "current_spice.out"


def solve_circuit(spice_file_contents, engine=spice.engine, raw=True):
    """
    Sends the spice-readable file to the spice engine which will run it and store the data in a temporary folder.
    Once the process is finished, it collects the data and returns it.

    If raw=False, this function will return only lines that start with a number. Depending of what was the output,
    the result might make sense or not. It is safer to ask for the raw data as it allows you to collect all the output
    from spice and deal with it as convenient, depending on the application.

    :param spice_file_contents: string formated as a spice-readable file contaning the design of the circuit and the instructions
    :param engine: the spice engine.
    :param raw: whether to produce the raw output or after some processing.
    :return: depending of the value of raw, this might be all the output of spice or just an array of data
    """

    spice.engine = engine
    if not spice.engine:
        raise SpiceSolverError("Spice solver not found.")

    with tempfile.TemporaryDirectory(prefix="tmp", suffix="_sc3NGSPICE") as working_directory:

        spice_file_path = os.path.join(working_directory, spice.input_file)
        spice_output_path = os.path.join(working_directory, spice.output_file)

        with open(spice_file_path, "w") as f:
            f.write(spice_file_contents)

        this_process = subprocess.Popen([spice.engine, '-b', spice_file_path, '-o', spice_output_path])
        this_process.wait()

        # this_process = subprocess.run([spice.engine, '-b', spice_file_path, '-o', spice_output_path])
        with open(spice_output_path, "r") as f:
            raw_results = f.read()

        if raw:
            # We return all the output
            return raw_results

        else:
            # We return just the lines starting with a number, which is OK in certain simple cases
            lines = raw_results.split("\n")
            data = []
            for line in lines:
                if len(line) == 0 or line[0] not in "1234567890.":
                    continue
                # print (line)

                i, *rest = line.split()
                # print (len(rest))
                data.append([float(element) for element in rest])

            return numpy.array(data).transpose()


def get_raw_from_spice(spice_file_contents, engine=spice):
    """ Dummy function included for backwards compatibility with solcore3 """

    results = solve_circuit(spice_file_contents, engine)

    return results


def send_to_spice(spice_file_contents, engine=spice):
    """ Dummy function included for backwards compatibility with solcore3 """

    results = solve_circuit(spice_file_contents, engine, raw=False)

    return results

if __name__ == '__main__':
    import matplotlib.pyplot as plt

    engine = spice.engine
    circuit = """Multiple dc sources
v1 1 0
r1 1 0 3.3k
.dc v1 0 24 1
.print dc i(v1)
.end
    """

    data = solve_circuit(circuit, engine=engine, raw=False)
    print(data)

    plt.plot(data[0], data[1])
    plt.show()


import numpy as np
from scipy.special import erfc
from scipy.interpolate import interp1d
from scipy.integrate import quad
import sys

from solcore import constants, si
from solcore.structure import Layer, Structure
from solcore.quantum_mechanics import schrodinger
from solcore.quantum_mechanics.kp_bulk import kp8x8_bulk
from solcore.quantum_mechanics.strain import strain_calculation_parameters
from solcore.absorption_calculator import adachi_alpha
from solcore.science_tracker import science_reference

q = constants.q
pi = constants.pi
h = constants.h
kb = constants.kb
m0 = constants.electron_mass
vacuum_permittivity = constants.vacuum_permittivity
c = constants.c


class QWunit(Structure):
    """ Assembles a group of layers as a quantum well, calculating its properties as a whole
    
    - It needs a minimum of 3 layers
    - There must be exactly one layer with the role = "well".
    - Top and bottom layers are assumed to be the barriers.
    - Anything not a barrier not a well will be considered as "interlayer".    
    
    Since the Poisson - Drift-Diffusion solver can not work with superlattices, there is no point of considering the solution of more than one QW.
    All QWs entering into the solver are independent, regardless of the width of the barriers.
    
    To account for the possible presence of nearby QWs, perdiodic boundary conditions can be used to solve the Schrodinger equation. 
    If barriers + interlayers are thick enought, this make no difference but if they are thin, it affects the number of confined levels. 
    It does not 'create' minibands, though, so the results with thin barriers must be used with caution.  
    
    """

    def __init__(self, *args, **kwargs):
        """ Initialises the QWunit, checking that it has all the necessary layers """
        super().__init__(*args, **kwargs)

        if len(self) < 3:
            print("ERROR creating a QWunit: a minimum of 3 layers is necessary to make a unit. Only {0} found. ".format(
                len(self)))
            sys.exit()

        self.QW_number = 0
        self.QW_width = 0

        self[0].__dict__["role"] = "barrier"
        self[-1].__dict__["role"] = "barrier"
        for index in range(1, len(self) - 1):

            if self[index].role in ["well", "qw", "QW", "Qw", "Well"]:
                self.QW_number += 1
                self.QW_width = self.QW_width + self[index].width
                self[index].role = "well"
            else:
                self[index].__dict__["role"] = "interlayer"

        # if self.QW_number != 1:
        #     print("ERROR creating a QWunit: There must be exactly one layer with the role = 'well'. {0} found. ".format(
        #         self.QW_number))
        #     sys.exit()

        if 'substrate' not in kwargs.keys():
            print("ERROR creating a QWunit: There must be a substrate defined with the keyword 'substrate'. ")
            sys.exit()

        # We update the structure of labels since that is what the quantum mechanics modules uses to identify the regions, not the role. 
        for index in range(len(self)):
            self.labels[index] = self[index].role

        self.T = kwargs['T'] if 'T' in kwargs.keys() else 298.0

    # ------
    def RecalculateKP(self, index, mode):
        """ Recalculates the band structure of the QW unit to account for modifications due to strain. 
        
        This is already done in the Schrodinger solver, but it does not produce that information as output. We do it so here.
        """
        if mode == 'kp8x8_bulk':
            kp_result = kp8x8_bulk(material=self[index].material, host_material=self.substrate, averaging_points=6)
            c, hh, lh, so, me, mhh, mlh, mso = kp_result

            # We recalculate the bandgap adn the electron afinity
            Ec = self[index].material.valence_band_offset + self[index].material.band_gap
            Eb = c - max(hh, lh)
            Xb = self[index].material.electron_affinity - c + Ec
        else:
            me = self[index].material.eff_mass_electron_Gamma * m0
            mhh = self[index].material.eff_mass_hh_z * m0
            mlh = self[index].material.eff_mass_lh_z * m0

            # Idem
            strain_parameters = strain_calculation_parameters(self.substrate, self[index].material, should_print=False)
            Eb = self[index].material.band_gap + strain_parameters.delta_Ec + max(
                strain_parameters.delta_Ehh, strain_parameters.delta_Elh)
            Xb = self[index].material.electron_affinity - strain_parameters.delta_Ec

        Ncc = 2 * (2 * pi * me * kb * self.T / h ** 2) ** 1.5
        Nvhh = 2 * (2 * pi * mhh * kb * self.T / h ** 2) ** 1.5
        Nvlh = 2 * (2 * pi * mlh * kb * self.T / h ** 2) ** 1.5

        return Eb, Xb, me, mhh, mlh, Ncc, Nvhh, Nvlh
        # ------

    def get_average_params(self, x, y):
        """ Calculates the average value of a parameter y in each layer of the structure.

        :param x: x values
        :param y: the parameter of interest at positions x
        :return: the average value of y in layer i
        """

        widths = np.array([0] + [layer.width for layer in self])
        cumul = np.cumsum(widths)

        aver = np.zeros(len(self))
        for i in range(len(self)):
            ind = np.searchsorted(x, cumul[i:i + 2])
            yy = y[ind[0]:ind[1] - 1]
            xx = x[ind[0]:ind[1] - 1]
            aver[i] = np.trapz(yy, xx) / (xx[-1] - xx[0])

        return aver

    def get_probability_per_layer(self, x, y):
        """ Calculates the average value of a parameter y in each layer of the structure.

        :param x: x values
        :param y: the parameter of interest at positions x
        :return: the average value of y in layer i
        """

        widths = np.array([0] + [layer.width for layer in self])
        cumul = np.cumsum(widths)

        aver = np.zeros(len(self))
        for i in range(len(self)):
            ind = np.searchsorted(x, cumul[i:i + 2])
            yy = y[ind[0]:ind[1] - 1]
            xx = x[ind[0]:ind[1] - 1]
            aver[i] = np.trapz(yy, xx)

        return aver

    def RecalculateBandEdges(self, mode, SR):
        """The effective bandgap and effective affinity of each layer depends on the energy level, e.g.: the minimum energy for electrons is not the band edge of the conduction band, but the ground level. In this function we modify that creating an effective electron afinity and band gaps for all layers in the QW.

        For the barriers, the electron afinity and band gap is the same than in bulk, modified by the kp calculation, if necesary."""

        eff_c_shift = np.where(SR["E"]["Ee"][0] > SR["potentials"]["Ve"], SR["E"]["Ee"][0],
                               SR["potentials"]["Ve"]) - SR["potentials"]["Ve"]

        max_VB = np.where(SR["potentials"]["Vhh"] > SR["potentials"]["Vlh"], SR["potentials"]["Vhh"],
                          SR["potentials"]["Vlh"])

        max_hole_level = max(SR["E"]["Ehh"][0], SR["E"]["Elh"][0])

        eff_h_shift = np.where(max_hole_level < max_VB, max_hole_level, max_VB) - max_VB

        shift_e = self.get_average_params(SR["x"], eff_c_shift)
        shift_h = self.get_average_params(SR["x"], eff_h_shift)

        new_values = []
        for index in range(len(self)):
            new_values.append(self.RecalculateKP(index, mode))

        for index in range(len(self)):
            Eb, Xb, me, mhh, mlh, Ncc, Nvhh, Nvlh = new_values[index]

            self[index].material.eff_mass_electron_Gamma = me
            self[index].material.eff_mass_hh_z = mhh
            self[index].material.eff_mass_lh_z = mlh

            self[index].material.__dict__["Ncc"] = 2 * (2 * pi * me * kb * self.T / h ** 2) ** 1.5
            self[index].material.__dict__["Nvhh"] = 2 * (2 * pi * mhh * kb * self.T / h ** 2) ** 1.5
            self[index].material.__dict__["Nvlh"] = 2 * (2 * pi * mlh * kb * self.T / h ** 2) ** 1.5

            self[index].band_gap = Eb
            self[index].electron_affinity = Xb

            self[index].__dict__["eff_band_gap"] = Eb + shift_e[index] - shift_h[index]
            self[index].__dict__["eff_electron_affinity"] = Xb - shift_e[index]

    # ------
    def RecalculateDensityOfStates(self, SR):
        """ Calculates the effective density of states for each layer in the QW. The general rule is:

        - Barriers have the bulk density of states
        - QW have ALL the density of states asociated with the confined states + bulk density of states above the barrier
        - Interlayers have only the bulk density of states above the barrier

        This simplification is similar to that in J. Nelson, M. Paxman, K. W. J. Barnham, J. S. Roberts, and C. Button, “Steady-state carrier escape from single quantum wells,” IEEE J. Quantum Electron., vol. 29, no. 6, pp. 1460–1468, 1993.

        From a physical porint of view, it can certainly be done much better"""

        science_reference("Density of state for QWs in the PDD solver",
                          "J. Nelson, M. Paxman, K. W. J. Barnham, J. S. Roberts, and C. Button, “Steady-state carrier escape from single quantum wells,” IEEE J. Quantum Electron., vol. 29, no. 6, pp. 1460–1468, 1993")

        # We use the average barrier electron afinity (considering only the barriers at the ends) as the electron afinity of the barrier
        Xb = (self[0].electron_affinity + self[-1].electron_affinity) / 2
        Egb = (self[0].band_gap + self[-1].band_gap) / 2

        b = q / (kb * self.T)

        e_prob = []
        hh_prob = []
        lh_prob = []
        for i in range(len(self.elevels)):
            e_prob.append(self.get_probability_per_layer(SR["x"], SR["wavefunctions"]['psi_e'][i] ** 2))
        for i in range(len(self.hhlevels)):
            hh_prob.append(self.get_probability_per_layer(SR["x"], SR["wavefunctions"]['psi_hh'][i] ** 2))
        for i in range(len(self.lhlevels)):
            lh_prob.append(self.get_probability_per_layer(SR["x"], SR["wavefunctions"]['psi_lh'][i] ** 2))

        for index in range(len(self)):
            Ncc = self[index].material.Ncc
            Nvhh = self[index].material.Nvhh
            Nvlh = self[index].material.Nvlh

            #   1- Contribution from bulk:
            Vconf = self[index].electron_affinity - Xb
            self[index].material.__dict__["Nc"] = Ncc * erfc(max(0, b * Vconf) ** 0.5)

            #   2- Contribution from confined levels
            Nqw = 0
            for i, eLevel in enumerate(self.elevels):
                Nqw = Nqw + np.exp(-b * eLevel) * e_prob[i][index]

            Nqw = 2. / self[index].width * (Ncc / 2.) ** (2. / 3.) * Nqw
            self[index].material.Nc = self[index].material.Nc + Nqw

            # For holes ------------------------------------------------------------------

            #   1- Contribution from bulk:
            Vconf = Xb + Egb - self[index].electron_affinity - self[index].band_gap
            self[index].material.__dict__["Nv"] = (Nvhh + Nvlh) * erfc(max(0, b * Vconf) ** 0.5)

            #   2- Contribution from heavy holes confined levels
            Nqw = 0
            for i, hhLevel in enumerate(self.hhlevels):
                Nqw = Nqw + (Nvhh / 2.) ** (2. / 3.) * np.exp(-b * hhLevel) * hh_prob[i][index]

                #   3- Contribution from light holes confined levels
            for i, lhLevel in enumerate(self.lhlevels):
                Nqw = Nqw + (Nvlh / 2.) ** (2. / 3.) * np.exp(-b * lhLevel) * lh_prob[i][index]

            Nqw = 2. / self[index].width * Nqw
            self[index].material.Nv = self[index].material.Nv + Nqw

    # ------
    def CalculateAbsorption(self, use_Adachi, SR):
        """ If required, this function calculates the absorption of the QW, putting together the absorption of the confined levels and the absorption of the bulk. As with the density of states, the rules are:

        - Barriers have the bulk absorption
        - Interlayers have the bulk absorption from the barrier energy and zero below that
        - Wells have the absorption of the confined levels below the barrier energy and of the bulk above it. The calculation is similar to: C. I. Cabrera, J. C. Rimada, J. P. Connolly, and L. Hernandez, “Modelling of GaAsP/InGaAs/GaAs strain-balanced multiple-quantum well solar cells,” J. Appl. Phys., vol. 113, no. 2, p. 024512, Jan. 2013."""

        science_reference("Absorption for QWs in the PDD solver",
                          "C. I. Cabrera, J. C. Rimada, J. P. Connolly, and L. Hernandez, “Modelling of GaAsP/InGaAs/GaAs strain-balanced multiple-quantum well solar cells,” J. Appl. Phys., vol. 113, no. 2, p. 024512, Jan. 2013.")

        edge = (self[0].band_gap + self[-1].band_gap) / 2 / q
        edge_index = np.abs(self.wl - 1240e-9 / edge).argmin()

        # We set the energy levels and absorption coefficient of each layer of the QW unit.
        self.absorption = 0 * self.wl
        for index in range(len(self)):
            if use_Adachi:
                try:
                    self[index].material.absorption = \
                        adachi_alpha.create_adachi_alpha(SolcoreMaterialToStr(self[index].material), T=self.T,
                                                         wl=self.wl)[
                            3]  # 0 = Energy, 1 = n, 2 = k, 3 = Absorption
                except:
                    self[index].material.absorption = self[index].material.alpha(self.wl)
                    # self[index].material.absorption[self.wl>1240e-9/sc.asUnit(self[index].material.band_gap, 'eV' )] = 0

            else:
                try:
                    self[index].material.absorption = self[index].material.alpha(self.wl)
                    # self[index].material.absorption[self.wl>1240e-9/sc.asUnit(self[index].material.band_gap, 'eV' )] = 0
                except:
                    print("Warning: Using Adachi calculation to estimate the absorption coefficient of layer: ",
                          self[index])
                    self[index].material.absorption = \
                        adachi_alpha.create_adachi_alpha(SolcoreMaterialToStr(self[index].material), T=self.T,
                                                         wl=self.wl)[
                            3]  # 0 = Energy, 1 = n, 2 = k, 3 = Absorption

            # self[index].material.absorption[ edge_index: ] = 0
            if self.labels[index] is "well":

                self[index].material.absorption = self[index].material.absorption - self[index].material.absorption[
                    edge_index]
                self[index].material.absorption[edge_index:] = 0
                self[index].material.absorption[self[index].material.absorption < 0] = 0
                self[index].material.absorption = self[index].material.absorption + SR["alpha"][1] * self[
                    index].width / self.QW_width

            elif self.labels[index] is "interlayer":
                self[index].material.absorption[edge_index:] = 0

            else:
                self[index].material.absorption[edge_index:] = 0

        self.absorption += self[index].material.absorption

    # ------
    def solve(self, Efield=0, WLsteps=(300e-9, 1100e-9, 801), wavelengths=None, periodic=True, filter_strength=0.0,
              blur=None, blurmode="left", offset=0, mode='kp8x8_bulk', use_Adachi=False, calculate_absorption=True,
              alpha_params=None, T=293):
        """ Solves the structure, calculating the energy levels, the absorption, etc. """

        if alpha_params == None:
            if wavelengths is None:
                self.wl = np.linspace(*WLsteps)
            else:
                self.wl = wavelengths

            E = 1240 / (self.wl * 1e9) * q

            alpha_params = {
                "well_width": self.QW_width,
                "theta": 0,
                "eps": 12.9 * vacuum_permittivity,
                "espace": E,
                "hwhm": si("4meV"),
                "dimensionality": 0.2,
                "line_shape": "Gauss"
            }
        else:
            self.wl = 1240 / (alpha_params["espace"] * 1e-9) * q

        SR, bands = schrodinger(self, mode=mode, periodic=True, calculate_absorption=calculate_absorption,
                                Efield=Efield, blur=None, blurmode=blurmode, alpha_params=alpha_params,
                                filter_strength=filter_strength)

        self.schrodinger = SR
        self.bands = bands

        self.__dict__["elevels"] = (SR["E"]["Ee"][:] - min(SR["potentials"]["Ve"])) / q
        self.__dict__["hhlevels"] = (max(SR["potentials"]["Vhh"]) - SR["E"]["Ehh"][:]) / q
        self.__dict__["lhlevels"] = (max(SR["potentials"]["Vlh"]) - SR["E"]["Elh"][:]) / q

        self.RecalculateBandEdges(mode, SR)
        self.RecalculateDensityOfStates(SR)
        if calculate_absorption:
            self.CalculateAbsorption(use_Adachi, SR)

        return SR

    def GetEffectiveQW(self, calculate_absorption=True, wavelengths=None, periodic=True, filter_strength=0.0, blur=None,
                       blurmode="left", mode='kp8x8_bulk', use_Adachi=False, alpha_params=None):
        """ Considers the device as a QW and solves its properties, including the modification of the bandeges due to strain, the efective mases and the absorption coefficient. The output is a list of layers made with materials with the effective properties after considering all these effects in addition to the quantum confinement.

        :param device: The device structure
        :param calculate_absorption: If absorption must be calculated
        :param WLsteps: wavelengths in which to calculate the absorption (input for np.linspace function)
        :param wavelengths: An array with the waveengths
        :param periodic: If it has to be assumed that the structure is perdiodic
        :param filter_strength:
        :param blur:
        :param blurmode:
        :param mode:
        :param use_Adachi:
        :param alpha_params:
        :return: A dictionary with the output of the Schrodinger solver.
        """
        print('Solving QW properties...')

        self.solve(calculate_absorption=calculate_absorption, wavelengths=wavelengths,
                   T=self.T, periodic=periodic, filter_strength=filter_strength, blur=blur,
                   blurmode=blurmode, mode=mode, use_Adachi=use_Adachi, alpha_params=alpha_params)

        for i in range(len(self)):
            self[i].material.band_gap = self[i].eff_band_gap
            self[i].material.electron_affinity = self[i].eff_electron_affinity
            self[i].material.ni = np.sqrt(
                self[i].material.Nc * self[i].material.Nv * np.exp(-self[i].eff_band_gap / (kb * self.T)))

        # Finally, we re-build a list of layers with the effective properties
        new_QW = []
        for i in range(len(self)):
            # We recover the composition and thickness
            layer_mat = self[i].material
            width = self[i].width

            # In the end, we convert the absorption coefficient in extinction coefficient
            kk = self[i].material.absorption * self.wl / 4 / np.pi
            layer_mat.k = interp1d(self.wl, kk, bounds_error=False, fill_value=(0, 0))

            # And the radiative recombination parameter
            inter = lambda E: 1.0 / layer_mat.ni ** 2 * 2 * pi / (h ** 3 * c ** 2) * layer_mat.n(
                1240e-9 / (E / q)) ** 2 * layer_mat.alphaE(E) * np.exp(-E / (kb * self.T)) * E ** 2
            Br = -np.trapz(np.nan_to_num(inter(1240e-9 / self.wl * q)), 1240e-9 / self.wl * q)
            layer_mat.radiative_recombination = Br

            # And add the layer to the list of layers
            new_QW.append(Layer(width=width, material=layer_mat))

        # As the QW might be actually a MQW, we repeat this as many times as needed
        new_QW = self.repeat * new_QW
        return new_QW


# ------
def SolcoreMaterialToStr(material_input):
    """Translate a solcore material composition into a string that is what the Adachi calculator needs."""

    material_string = material_input.__str__().strip('<>').split(" ")
    material_name = material_string[0].strip("'")
    composition = {'material': material_name}
    if len(material_name) > 4:
        material_composition = material_string[2].split("=")
        for i, comp in enumerate(material_composition):
            if comp in material_name:
                composition['element'] = material_composition[i]
                composition['fraction'] = float(material_composition[i + 1])

    output = "%s {'%s':%s}" % (composition['material'], composition['element'], composition['fraction'])

    return output



"""
This is completely based on Pablo R. Mier's yabox package, with modifications to slightly change what is returned by
DE.solve to be able to visualize the convergence of the DE without having to use de.iterator, and just return
a list of optimized parameters rather than a reshaped array.

All credit of the algorithm, testing, etc. goes to Pablo R. Mier. For more details, visit:
- https://github.com/pablormier/yabox
- https://pablormier.github.io/2017/09/05/a-tutorial-on-differential-evolution-with-python/
"""

from yabox.algorithms.de import PDEIterator
from yabox.algorithms.de import DE as DE_yabox
import numpy as np


class DE(DE_yabox):
    def solve(self, show_progress=False):
        best_pop_evo = []
        best_fitn_evo = []
        mean_fitn_evo = []
        if show_progress:
            from tqdm import tqdm
            iterator = tqdm(self.iterator(), total=self.maxiters, desc='Optimizing ({0})'.format(self.name))
        else:
            iterator = self.iterator()
        for step in iterator:
            idx = step.best_idx
            P = step.population

            fitness = step.fitness
            if step.iteration > self.maxiters:
                if show_progress:
                    iterator.n = self.maxiters
                    iterator.refresh()
                    iterator.close()
                break
            best_pop_evo.append(P[idx])
            best_fitn_evo.append(fitness[idx])
            mean_fitn_evo.append(np.mean(fitness))
        return self.denormalize(P[idx]), fitness[idx], \
               self.denormalize(best_pop_evo[self.popsize-1::self.popsize]), \
               np.array(best_fitn_evo[self.popsize-1::self.popsize]), \
               np.array(mean_fitn_evo[self.popsize-1::self.popsize])



class PDE(DE):
    def __init__(self, fobj, bounds, mutation=(0.5, 1.0), crossover=0.7, maxiters=1000,
                 self_adaptive=False, popsize=None, seed=None, processes=None, chunksize=None):
        super().__init__(fobj, bounds, mutation, crossover, maxiters, self_adaptive, popsize, seed)
        from multiprocessing import Pool
        self.processes = processes
        self.chunksize = chunksize
        self.name = 'Parallel DE'
        self.pool = None
        if processes is None or processes > 0:
            self.pool = Pool(processes=self.processes)

    def iterator(self):
        it = PDEIterator(self)
        try:
            for data in it:
                yield data
        finally:
            if self.pool is not None:
                self.pool.terminate()

    def evaluate_denormalized(self, PD):
        if self.pool is not None:
            return list(self.pool.map(self.fobj, PD, chunksize=self.chunksize))
        else:
            return super().evaluate_denormalized(PD)



from pytest import approx, mark


@mark.flaky(reruns=5)
def test_DE():
    from yabox.problems import Ackley
    from solcore.optimization import DE

    problem = Ackley()

    algorithm = DE(problem, problem.bounds)

    result = algorithm.solve()

    assert result[0] == approx(0.0)
    assert result[1] == approx(0.0)
    assert result[2][-1] == approx(result[0])
    assert result[3][-1] == approx(0.0)
    assert result[4][-1] == approx(0.0)


@mark.flaky(reruns=5)
def test_PDE():
    from yabox.problems import Ackley
    from solcore.optimization import PDE

    problem = Ackley()

    algorithm = PDE(problem, problem.bounds)

    result = algorithm.solve()

    assert result[0] == approx(0.0)
    assert result[1] == approx(0.0)
    assert result[2][-1] == approx(result[0])
    assert result[3][-1] == approx(0.0)
    assert result[4][-1] == approx(0.0)



""" Absorption calculator related tests
"""
from pytest import approx

from solcore import material, si
from solcore.structure import Structure, Layer
from solcore.absorption_calculator import create_adachi_alpha
from solcore.absorption_calculator import (
    calculate_rat,
    calculate_ellipsometry,
    calculate_absorption_profile,
)
from solcore.state import State
from solcore.solar_cell_solver import prepare_solar_cell
from solcore.absorption_calculator.dielectric_constant_models import (
    DielectricConstantModel,
    Drude,
)

from solcore.solar_cell_solver import solar_cell_solver
from solcore.solar_cell import SolarCell
from solcore.material_system import create_new_material
from solcore.optics import solve_tmm

import os
from pathlib import Path

import numpy as np


def test_adachi_absorption():
    material_name = "InGaAs"
    solcore_material = material(material_name)(T=300, In=0.1)

    E, nn, kk, adachi_alpha_data = create_adachi_alpha(solcore_material, T=300)

    E /= 1.6e-19

    idx = np.argmin(abs(E - 2))

    out = [E[idx], nn[idx], kk[idx]]

    data = [2.0003267755918639, 3.6288603412514253, 0.30754355994545018]

    assert all([d == approx(o) for d, o in zip(data, out)])


def test_TMM_rat():
    GaAs = material("GaAs")(T=300)

    my_structure = Structure([Layer(si(3000, "nm"), material=GaAs)])

    wavelength = np.linspace(450, 1100, 300)
    idx = np.argmin(abs(wavelength - 800))

    out = calculate_rat(
        my_structure, wavelength, coherent=True, no_back_reflexion=False
    )
    out = (out["R"][idx], out["A"][idx], out["T"][idx])

    data = (0.33328918841743332, 0.65996607786373396, 0.0067447337188326914)

    assert all([d == approx(o) for d, o in zip(data, out)])


def test_TMM_absorption_profile():
    GaAs = material("GaAs")(T=300)

    my_structure = Structure(
        [Layer(si(3000, "nm"), material=GaAs), Layer(si(300, "um"), material=GaAs)]
    )

    out = calculate_absorption_profile(
        my_structure, np.array([800]), z_limit=3000, steps_size=20
    )

    data_path = Path(__file__).parent / "data" / "absorption_profile.txt"
    data = tuple(np.loadtxt(data_path))

    assert all([d == approx(o) for d, o in zip(data, out["absorption"][0])])


def test_TMM_ellipsometry():
    GaAs = material("GaAs")(T=300)

    my_structure = Structure(
        [Layer(si(3000, "nm"), material=GaAs), Layer(si(300, "um"), material=GaAs)]
    )

    wavelength = np.linspace(450, 1100, 300)
    idx = np.argmin(abs(wavelength - 800))

    angles = [60, 65, 70]
    out = calculate_ellipsometry(my_structure, wavelength, angle=angles)

    data = (
        22.2849089096,
        181.488417672,
        16.4604621886,
        182.277656469,
        9.10132195668,
        184.509752582,
    )

    for i in range(len(angles)):
        assert data[2 * i] == approx(out["psi"][idx, i])
        assert data[2 * i + 1] == approx(out["Delta"][idx, i])


def test_TMM_dielectric_model():
    drud = Drude(An=24.317000, Brn=0.125740)

    model = DielectricConstantModel(e_inf=3.4837, oscillators=[drud])

    wavelength = 2 * np.logspace(3, 4, 10)
    n = model.n_and_k(wavelength)

    data = (
        0.37377710 + 2.0726883j,
        0.53555835 + 3.03640188j,
        0.81383835 + 4.12781828j,
        1.25563998 + 5.39395751j,
        1.92396319 + 6.8504966j,
        2.88306008 + 8.48061105j,
        4.17692618 + 10.23820188j,
        5.81068443 + 12.06785126j,
        7.75184851 + 13.93546615j,
        9.95389660 + 15.84816722j,
    )

    assert all([d == approx(o) for d, o in zip(data, n)])


def test_sopra_absorption():
    from solcore.absorption_calculator import sopra_database

    # Import material constant data for Gallium Arsenide :: Do this by placing the
    # material name as the sole argument...
    SOPRA_Material = sopra_database("GaAs")

    # Can also load alpha data...
    GaAs_alpha = SOPRA_Material.load_alpha()

    out = GaAs_alpha[1][10]

    data = 163666134.03339368

    assert data == approx(out)


def test_substrate_presence_A():
    wavelength = np.linspace(300, 800, 3) * 1e-9

    GaAs = material("GaAs")(T=300)

    my_structure = SolarCell([Layer(si(700, "nm"), material=GaAs)], substrate=GaAs)

    solar_cell_solver(
        my_structure,
        "optics",
        user_options={
            "wavelength": wavelength,
            "optics_method": "TMM",
            "no_back_reflexion": False,
        },
    )

    z_pos = np.linspace(0, my_structure.width, 10)

    A_subs = my_structure[0].layer_absorption

    my_structure = SolarCell([Layer(si(700, "nm"), material=GaAs)])

    solar_cell_solver(
        my_structure,
        "optics",
        user_options={
            "wavelength": wavelength,
            "optics_method": "TMM",
            "no_back_reflexion": False,
        },
    )

    A_nosubs = my_structure[0].layer_absorption

    A = np.vstack((A_subs, A_nosubs))

    A_data = np.array(
        [[0.56610281, 0.62692985, 0.41923175], [0.56610281, 0.62711355, 0.37837737]]
    )

    assert all([d == approx(o) for d, o in zip(A, A_data)])


def test_BL_correction():

    wl = np.linspace(800, 950, 4) * 1e-9

    GaAs = material("GaAs")()

    thick_cell = SolarCell([Layer(material=GaAs, width=si("20um"))])

    opts = State()
    opts.position = None
    prepare_solar_cell(thick_cell, opts)
    position = np.arange(0, thick_cell.width, 1e-9)
    opts.position = position
    opts.recalculate_absorption = True
    opts.no_back_reflexion = False

    opts.BL_correction = False
    opts.wavelength = wl
    solve_tmm(thick_cell, opts)

    no_corr = thick_cell.absorbed

    opts.BL_correction = True

    solve_tmm(thick_cell, opts)

    with_corr = thick_cell.absorbed

    assert with_corr == approx(
        np.array([6.71457872e-01, 6.75496354e-01, 2.09738887e-01, 0])
    )
    assert no_corr == approx(
        np.array([6.71457872e-01, 6.75496071e-01, 2.82306407e-01, 0])
    )


def test_substrate_presence_profile():
    wavelength = np.linspace(300, 800, 3) * 1e-9

    GaAs = material("GaAs")(T=300)

    my_structure = SolarCell([Layer(si(700, "nm"), material=GaAs)], substrate=GaAs)

    solar_cell_solver(
        my_structure,
        "optics",
        user_options={
            "wavelength": wavelength,
            "optics_method": "TMM",
            "no_back_reflection": False,
        },
    )

    z_pos = np.linspace(0, my_structure.width, 10)

    profile_subs = my_structure[0].absorbed(z_pos)

    my_structure = SolarCell([Layer(si(700, "nm"), material=GaAs)])

    solar_cell_solver(
        my_structure,
        "optics",
        user_options={
            "wavelength": wavelength,
            "optics_method": "TMM",
            "no_back_reflection": False,
        },
    )

    profile_nosubs = my_structure[0].absorbed(z_pos)
    profile = np.vstack((profile_subs, profile_nosubs))

    data_path = Path(__file__).parent / "data" / "substrate_presence_profile.csv"
    expected = np.loadtxt(data_path, delimiter=",")

    assert profile.shape == expected.shape
    assert profile == approx(expected)


def test_inc_coh_tmm():
    GaInP = material("GaInP")(In=0.5)
    GaAs = material("GaAs")()
    Ge = material("Ge")()

    optical_struct = SolarCell(
        [
            Layer(material=GaInP, width=si("5000nm")),
            Layer(material=GaAs, width=si("200nm")),
            Layer(material=GaAs, width=si("5um")),
            Layer(material=Ge, width=si("50um")),
        ]
    )

    wl = np.linspace(400, 1200, 5) * 1e-9

    options = State()
    options.wavelength = wl
    options.optics_method = "TMM"
    options.no_back_reflection = False
    options.BL_correction = True
    options.recalculate_absorption = True

    c_list = [
        ["c", "c", "c", "c"],
        ["c", "c", "c", "i"],
        ["c", "i", "i", "c"],
        ["i", "i", "i", "i"],
    ]

    results = []
    for i1, cl in enumerate(c_list):
        options.coherency_list = cl
        solar_cell_solver(optical_struct, "optics", options)
        results.append(optical_struct.absorbed)

    A_calc = np.stack(results)
    A_data = np.array(
        [
            [0.5742503, 0.67956899, 0.73481184, 0.725372, 0.76792856],
            [0.5742503, 0.67956899, 0.73481184, 0.725372, 0.76792856],
            [0.5742503, 0.67956899, 0.73474943, 0.70493469, 0.70361194],
            [0.5742503, 0.67956899, 0.70927724, 0.71509221, 0.71592772]
        ]
    )
    assert A_calc == approx(A_data)


def test_define_material():
    from solcore import ParameterSystem, MaterialSystem
    this_dir = os.path.split(__file__)[0]
    create_new_material(
        "SiGeSn",
        os.path.join(this_dir, "data", "SiGeSn_n.txt"),
        os.path.join(this_dir, "data", "SiGeSn_k.txt"),
        os.path.join(this_dir, "data", "SiGeSn_params.txt"),
        overwrite=True,
    )
    assert "SiGeSn" in ParameterSystem().database.sections()
    assert "SiGeSn".lower() in MaterialSystem().sources

    SiGeSn = material("SiGeSn")()
    assert SiGeSn.n(400e-9) == approx(4.175308391752484)
    assert SiGeSn.k(400e-9) == approx(2.3037424963866306)



import numpy as np
import tmm
from solcore.absorption_calculator import OptiStack, RCWASolverError
from joblib import Parallel, delayed

try:
    import S4
except ModuleNotFoundError:
    raise RCWASolverError


DEFAULT_OPTIONS = dict(LatticeTruncation='Circular',
                        DiscretizedEpsilon=False,
                        DiscretizationResolution=8,
                        PolarizationDecomposition=False,
                        PolarizationBasis='Default',
                        LanczosSmoothing=False,
                        SubpixelSmoothing=False,
                        ConserveMemory=False,
                        WeismannFormulation=False)


def calculate_rat_rcwa(structure, size, orders, wavelength, incidence, substrate, theta=0, phi=0, pol='u',
                       parallel=False, n_jobs=-1, user_options=None):
    """Calculates the reflected, absorbed and transmitted intensity of the structure for the wavelengths and angles
    defined using an RCWA method implemented using the S4 package.
    This function is analogous to calculate_rat from the transfer_matrix module.

    :param structure: A Solcore Structure object with layers and materials or a OptiStack object.
    :param size: a tuple of 2-D vectors in the format ((ux, uy), (vx, vy)) giving the x and y components of the lattice unit vectors in nm.
    :param orders: number of orders to retain in the RCWA calculations.
    :param wavelength: Wavelengths (in nm) in which calculate the data.
    :param incidence: a Solcore material describing the semi-infinite incidence medium
    :param substrate: a Solcore material describing the semi-infinite transmission medium
    :param theta: polar incidence angle (in degrees) of the incident light. Default: 0 (normal incidence)
    :param phi: azimuthal incidence angle (in degrees). Default: 0
    :param pol: Polarisation of the light: 's', 'p' or 'u'. Default: 'u' (unpolarised).
    :param parallel: whether or not to execute calculation in parallel (over wavelengths), True or False. Default is False
    :param n_jobs: the 'n_jobs' argument passed to Parallel from the joblib package. If set to -1, all available CPUs are used,
            if set to 1 no parallel computing is executed. The number of CPUs used is given by n_cpus + 1 + n_jobs. Default is -1.
    :param user_options: dictionary of options for S4. The list of possible entries and their values is:

            * LatticeTruncation: 'Circular' or 'Parallelogramic' (default 'Circular')
            * DiscretizedEpsilon: True or False (default False)
            * DiscretizationResolution: integer (default value 8)
            * PolarizationDecomposition: True or False (default False)
            * PolarizationBasis: 'Default' or 'Normal' or 'Jones' (default 'Default')
            * LanczosSmoothing: True or False (default False)
            * SubpixelSmoothing: True or False (default False)
            * ConserveMemory: True or False (default False)
            * WeismannFormulation: True or False (default False)

        Further information on the function of these options can be found in the S4 Python API documentation. If no
        options are provided, those from DEFAULT_OPTIONS are used.

    :return: A dictionary with the R, total A and T at the specified wavelengths and angle. A_pol lists total absorption
       at s and p polarizations if pol = 'u' was specified (this information is needed for the absorption profile calculation).
       Otherwise, A_pol is the same as A.
    """
    num_wl = len(wavelength)

    # write a separate function that makes the OptiStack structure into an S4 object, defined materials etc.
    # Materials for the shapes need to be defined before you can do .SetRegion

    geom_list = [layer.geometry for layer in structure]
    geom_list.insert(0, {})  # incidence medium
    geom_list.append({})  # transmission medium

    shape_mats, geom_list_str = necessary_materials(geom_list)

    shapes_oc = np.zeros((len(wavelength), len(shape_mats)), dtype=complex)

    for i1, x in enumerate(shape_mats):
        shapes_oc[:, i1] = (x.n(wavelength/1e9) + 1j * x.k(wavelength/1e9)) ** 2

    stack_OS = OptiStack(structure, bo_back_reflection=False, substrate=substrate)
    widths = stack_OS.get_widths()
    layers_oc = np.zeros((len(wavelength/1e9), len(structure) + 2), dtype=complex)

    if incidence == None:
        layers_oc[:, 0] = 1
    else:
        layers_oc[:, 0] = (incidence.n(wavelength/1e9)) ** 2  # + 1j*incidence.k(wavelengths))**2


    if substrate == None:
        layers_oc[:, -1] = 1
    else:
        layers_oc[:, -1] = (substrate.n(wavelength/1e9) + 1j * substrate.k(wavelength/1e9)) ** 2

    for i1, x in enumerate(structure):
        layers_oc[:, i1 + 1] = (x.material.n(wavelength/1e9) + 1j * x.material.k(wavelength/1e9)) ** 2

    shapes_names = [str(x) for x in shape_mats]

    rcwa_options = DEFAULT_OPTIONS
    if user_options is not None:
        rcwa_options.update(user_options)

    if parallel:
        allres = Parallel(n_jobs=n_jobs)(delayed(RCWA_wl)
                                                         (wavelength[i1], geom_list_str,
                                                          layers_oc[i1], shapes_oc[i1],
                                                          shapes_names, pol,
                                                          theta, phi,
                                                          widths, size,
                                                          orders, rcwa_options)
                                                         for i1 in range(num_wl))

    else:
        allres = [
            RCWA_wl(wavelength[i1], geom_list_str, layers_oc[i1], shapes_oc[i1],
                         shapes_names, pol, theta, phi,
                         widths, size,
                         orders, rcwa_options)
            for i1 in range(num_wl)]

    R = np.stack([item[0] for item in allres])
    T = np.stack([item[1] for item in allres])
    A_mat = np.stack([item[2] for item in allres])
    if pol == 'u':
        A_per_layer = np.mean(A_mat, axis=1)
        A_pol = np.sum(A_mat, 2)
    else:
        A_per_layer = A_mat
        A_pol = np.sum(A_per_layer, 1)


    output = {'R': R, 'A': np.sum(A_per_layer, 1), 'T': T,
              'A_per_layer': A_per_layer, 'A_pol': A_pol}

    return output


def rcwa_rat(S, n_layers, theta, n_inc):
    """
    Calculates total reflection and transmission and reflection and transmission by diffraction order, at a single wavelength

    :param S: S4 simulation object
    :param n_layers: the number of layers in the structure, including the incidence and transmission medium
    :param theta: incident polar angle in degrees
    :param n_inc: (real part of) the refractive index of the incidence medium
    :return: dictionary with total R, total T, and lists of R and T per order
    """
    # TODO: check normalization of pfbo values. Also, theta correction should happen inside here rather than externally
    below = 'layer_' + str(n_layers)  # identify which layer is the transmission medium

    R = -n_inc*np.real(S.GetPowerFlux('layer_1')[1])/np.cos(theta*np.pi/180)  # GetPowerFlux gives forward & backward Poynting vector, so sum to get power flux


    R_pfbo = -np.array(S.GetPowerFluxByOrder('layer_1'))[:,1] # real part of backwards power flow. Not normalised correctly.

    T = n_inc*sum(S.GetPowerFlux(below))/np.cos(theta*np.pi/180)

    T_pfbo = n_inc*np.real(np.sum(np.array(S.GetPowerFluxByOrder(below)), 1))/np.cos(theta*np.pi/180)

    return {'R': np.real(R), 'T': np.real(T), 'R_pfbo': R_pfbo, 'T_pfbo': T_pfbo}


def initialise_S(size, orders, geom_list, mats_oc, shapes_oc, shape_mats, widths, options):
    """Makes an S4 simulation object using S4.New() according to the user's specified structure and options.

    :param size: a tuple of 2-D vectors in the format ((ux, uy), (vx, vy)) giving the x and y components of the lattice unit vectors in nm.
    :param orders: number of Fourier orders to be retained in the RCWA/FMM calculation
    :param: geom_list: list containing shape information fpr each layer. Format is a list of lists; entries of inner lists are one dictionary
            per shape containing the shape information relevant for that shape.
    :param: mats_oc: complex dielectric constant for each of the base layers (including incidence and transmission media)
    :param: shapes_oc: complex dielectric constants for materials in the shapes inside layers
    :param: shape_mats: names of the shape materials
    :param: widths: widths of the layers, including the semi-infinite incidence and transmission medium. In nm.
    :param: options: S4 options (dictionary)

    :return: S4 simulation object
    """
    S = S4.New(size, orders)

    S.SetOptions(
        LatticeTruncation = options['LatticeTruncation'],
        DiscretizedEpsilon = options['DiscretizedEpsilon'],
        DiscretizationResolution = options['DiscretizationResolution'],
        PolarizationDecomposition = options['PolarizationDecomposition'],
        PolarizationBasis = options['PolarizationBasis'],
        LanczosSmoothing = options['LanczosSmoothing'],
        SubpixelSmoothing = options['SubpixelSmoothing'],
        ConserveMemory = options['ConserveMemory'],
        WeismannFormulation = options['WeismannFormulation']
    )


    for i1, shape in enumerate(shapes_oc):
        S.SetMaterial('shape_mat_' + str(i1 + 1), shape)

    for i1 in range(len(widths)):
        S.SetMaterial('layer_' + str(i1 + 1), mats_oc[i1])

    for i1, width in enumerate(widths):  # set base layers
        layer_name = 'layer_' + str(i1 + 1)
        #print(layer_name)
        if width == float('Inf'):
            #print('zero width')
            S.AddLayer(layer_name, 0, layer_name)  # Solcore4 has incidence and transmission media widths set to Inf;
            # in S4 they have zero width
        else:
            S.AddLayer(layer_name, width, layer_name)

        geometry = geom_list[i1]

        if bool(geometry):
            for shape in geometry:

                mat_name = 'shape_mat_' + str(shape_mats.index(str(shape['mat'])) + 1)
                #print(str(shape['mat']), mat_name)
                if shape['type'] == 'circle':
                    S.SetRegionCircle(layer_name, mat_name, shape['center'], shape['radius'])
                elif shape['type'] == 'ellipse':
                    S.SetRegionEllipse(layer_name, mat_name, shape['center'], shape['angle'], shape['halfwidths'])
                elif shape['type'] == 'rectangle':
                    #print('rect')
                    S.SetRegionRectangle(layer_name, mat_name, shape['center'], shape['angle'], shape['halfwidths'])
                elif shape['type'] == 'polygon':
                    S.SetRegionPolygon(layer_name, mat_name, shape['center'], shape['angle'], shape['vertices'])

    return S


def necessary_materials(geom_list):
    """Checks which materials are needed to define the shapes in the layers (so not the main layer materials), and
    returns a version of the geom_list object which doesn't contain Solcore materials (to avoid pickling issues).

    :param geom_list: list of the geometry attributes of every layer in the structure. This is a list of lists;
            each lists can contain any number of dictionaries (including 0) describing a shape in that layer
    :return: a list of the necessary Solcore materials for the calculation, and an object which is exactly like
            geom_list except the Solcore material objects have been replaced with strings. This is to avoid pickling
            issues when running calculations in parallel.
    """
    shape_mats = []
    geom_list_str = [None] * len(geom_list)
    for i1, geom in enumerate(geom_list):
        if bool(geom):
            shape_mats.append([x['mat'] for x in geom])
            geom_list_str[i1] =[{} for _ in range(len(geom))]
            for i2, g in enumerate(geom):
                for item in g.keys():
                    if item != 'mat':
                        geom_list_str[i1][i2][item] = g[item]
                    else:
                        geom_list_str[i1][i2][item] = str(g[item])

    return list(set([val for sublist in shape_mats for val in sublist])), geom_list_str


def update_epsilon(S, stack_OS, shape_mats_OS, wl):
    """Updates the dielectric constants in the S4 simulation object to the relevant values at the desired wavelength.

    :param S: S4 simulation object
    :param stack_OS: OptiStack for the main layers in the structures
    :param shape_mats_OS: OptiStack containing layers made up of the materials used in the shapes within the main layers
            (the structure is not physically meaningful but is used to store the optical constant information)
    :param wl: wavelength in nm
    :return: the S4 simulation object with updated parameters
    """
    for i1 in range(len(stack_OS.get_widths())):
        S.SetMaterial('layer_' + str(i1 + 1), stack_OS.get_indices(wl)[i1] ** 2)
    for i1 in range(len(shape_mats_OS.widths)):  # initialise the materials needed for all the shapes in S4
        S.SetMaterial('shape_mat_' + str(i1 + 1), shape_mats_OS.get_indices(wl)[i1 + 1] ** 2)

    return S

def calculate_absorption_profile_rcwa(structure, size, orders, wavelength, rat_output_A,
                                      z_limit=None, steps_size=2, dist=None, theta=0, phi=0, pol='u', incidence=None,
                                      substrate=None,
                                      parallel=False, n_jobs=-1, user_options=None):
    """It calculates the absorbed energy density within the material. Integrating this absorption profile in the whole stack gives the same result as the absorption obtained with
    calculate_rat as long as the spatial mesh is fine enough. If the structure is
    very thick and the mesh not thin enough, the calculation might diverge at short wavelengths.
    This function is analogous to calculate_absorption_profile from the transfer_matrix module.

    :param structure: A Solcore structure with layers and materials.
    :param size: a tuple of 2-D vectors in the format ((ux, uy), (vx, vy)) giving the x and y components of the lattice unit vectors in nm.
    :param orders: number of orders to retain in the RCWA calculations.
    :param wavelength: Wavelengths (in nm) in which calculate the data.
    :param rat_output: 'A_pol' (polarization-dependent layer absorption) output from calculate_rat_rcwa
    :param z_limit: Maximum value in the z direction at which to calculate depth-dependent absorption (nm)
    :param steps_size: if the dist is not specified, the step size in nm to use in the depth-dependent calculation
    :param dist: the positions (in nm) at which to calculate depth-dependent absorption
    :param theta: polar incidence angle (in degrees) of the incident light. Default: 0 (normal incidence)
    :param phi: azimuthal incidence angle in degrees. Default: 0
    :param pol: Polarisation of the light: 's', 'p' or 'u'. Default: 'u' (unpolarised).
    :param substrate: semi-infinite transmission medium
    :return: A dictionary containing the positions (in nm) and a 2D array with the absorption in the structure as a
        function of the position and the wavelength.
    """
    num_wl = len(wavelength)

    geom_list = [layer.geometry for layer in structure]
    geom_list.insert(0, {})  # incidence medium
    geom_list.append({})  # transmission medium
    # write a separate function that makes the OptiStack structure into an S4 object, defined materials etc.
    # write a separate function that makes the OptiStack structure into an S4 object, defined materials etc.

    ## Materials for the shapes need to be defined before you can do .SetRegion
    shape_mats, geom_list_str = necessary_materials(geom_list)

    shapes_oc = np.zeros((len(wavelength), len(shape_mats)), dtype=complex)

    for i1, x in enumerate(shape_mats):
        shapes_oc[:, i1] = (x.n(wavelength/1e9) + 1j * x.k(wavelength/1e9)) ** 2

    stack_OS = OptiStack(structure, bo_back_reflection=False, substrate=substrate)
    widths = stack_OS.get_widths()
    layers_oc = np.zeros((len(wavelength/1e9), len(structure) + 2), dtype=complex)

    if incidence == None:
        layers_oc[:, 0] = 1
    else:
        layers_oc[:, 0] = (incidence.n(wavelength/1e9)) ** 2  # + 1j*incidence.k(wavelengths))**2

    if substrate == None:
        layers_oc[:, -1] = 1
    else:
        layers_oc[:, -1] = (substrate.n(wavelength / 1e9) + 1j * substrate.k(wavelength / 1e9)) ** 2

    for i1, x in enumerate(structure):
        layers_oc[:, i1 + 1] = (x.material.n(wavelength/1e9) + 1j * x.material.k(wavelength/1e9)) ** 2

    shapes_names = [str(x) for x in shape_mats]
    rcwa_options = DEFAULT_OPTIONS

    if user_options is not None:
        rcwa_options.update(user_options)

    if dist is None:
        if z_limit is None:
            stack = OptiStack(structure)
            z_limit = np.sum(np.array(stack.widths))
        dist = np.arange(0, z_limit, steps_size)

    output = {'position': dist, 'absorption': np.zeros((num_wl, len(dist)))}

    if parallel:
        allres = Parallel(n_jobs=n_jobs)(delayed(RCWA_wl_prof)
                                                         (wavelength[i1], rat_output_A[i1],
                                                          dist,
                                                          geom_list_str,
                                                          layers_oc[i1], shapes_oc[i1],
                                                          shapes_names, pol,
                                                          theta, phi,
                                                          widths, size,
                                                          orders, rcwa_options)
                                                         for i1 in range(num_wl))

    else:
        allres = [
            RCWA_wl_prof(wavelength[i1], rat_output_A[i1],
                                                          dist,
                                                          geom_list_str,
                                                          layers_oc[i1], shapes_oc[i1],
                                                          shapes_names, pol,
                                                          theta, phi,
                                                          widths, size,
                                                          orders, rcwa_options)
            for i1 in range(num_wl)]

    profile = np.stack(allres)
    output['absorption'] = profile

    return output


def rcwa_position_resolved(S, layer, depth, A, theta, n_inc):
    """Calculate the derivative of the absorbed power with respect to depth in a layer.

    :param S: S4 simulation object
    :param layer: layer index (layer 1 is the incidence medium)
    :param depth: depth in the layer in nm
    :param A: the total power absorbed in the layer
    :param theta: polar angle of the light in the incidence medium
    :param n_inc: (real part of) the refractive index of the incidence medium
    :return: the derivative of the absorbed power with depth in the structure
    """
    if A > 0:
        delta = 1e-9
        power_difference = np.real(
            sum(S.GetPowerFlux(layer, depth - delta)) - sum(S.GetPowerFlux(layer, depth + delta)))
        return n_inc * power_difference / (2 * delta * np.cos(theta*np.pi/180))  # absorbed energy density normalised to total absorption
    else:
        return 0

def rcwa_absorption_per_layer(S, n_layers, theta, n_inc):
    """Calculates the fraction of the incident power absorbed in each layer of the structure.

    :param S: S4 simulation object
    :param n_layers: number of layers in the structure, including the incidence and transmission medium
    :param theta: polar angle of the light in the incidence medium
    :param n_inc: (real part of) the refractive index of the incidence medium
    :return:
    """
    A = np.empty(n_layers-2)
    for i1, layer in enumerate(np.arange(n_layers-2)+2):
        A[i1] = np.real(sum(S.GetPowerFlux('layer_' + str(layer))) -
                        sum(S.GetPowerFlux('layer_' + str(layer+1))))

    A = n_inc*np.array([x if x > 0 else 0 for x in A])/np.cos(theta*np.pi/180)

    return A

def RCWA_wl(wl, geom_list, l_oc, s_oc, s_names, pol, theta, phi, widths, size, orders, rcwa_options):

    """
    Calculates the reflection, absorption per layer and transmission using RCWA per wavelength. Called from calculate_rat_rcwa,
    can be used in parallel with joblib.

    :param wl: wavelength in nm
    :param geom_list: list containing shape information fpr each layer. Format is a list of lists; entries of inner lists are one dictionary
        per shape containing the shape information relevant for that shape.
    :param l_oc: list of optical constants of the main layers in the stack
    :param s_oc: list of optical constants for the shapes inside the layers
    :param s_names: list of names of the shapes
    :param pol: polarization ('s', 'p' or 'u')
    :param theta: polar incidence angle in degrees
    :param phi: azimuthal angle (clockwise from y-axis) in degrees)
    :param widths: layer widths in nm. Incidence and transmission medium have 0 thickness.
    :param size: a tuple of 2-D vectors in the format ((ux, uy), (vx, vy)) giving the x and y components of the lattice unit vectors in nm.
    :param orders: number of orders to retain in the RCWA calculations.
    :param rcwa_options: S4 options (dictionary)
    :return: R, T and A per layer at the specified wavelength
    """

    def vs_pol(s, p):
        S.SetExcitationPlanewave((theta, phi), s, p, 0)
        S.SetFrequency(1 / wl)
        out = rcwa_rat(S, len(widths), theta, n_inc)
        R = out['R']
        T = out['T']
        A_layer = rcwa_absorption_per_layer(S, len(widths), theta, n_inc)
        return R, T, A_layer

    S = initialise_S(size, orders, geom_list, l_oc, s_oc, s_names, widths, rcwa_options)

    n_inc = np.real(np.sqrt(l_oc[0]))

    if len(pol) == 2:

        R, T, A_layer = vs_pol(pol[0], pol[1])

    else:
        if pol in 'sp':
            R, T, A_layer = vs_pol(int(pol == "s"), int(pol == "p"))

        else:
            R_s, T_s, A_layer_s = vs_pol(1, 0)
            R_p, T_p, A_layer_p = vs_pol(0, 1)
            R = (R_s + R_p) / 2
            T = (T_s + T_p) / 2
            A_layer = np.stack([A_layer_s, A_layer_p])

    return R, T, A_layer


def RCWA_wl_prof(wl, A, dist, geom_list, layers_oc, shapes_oc, s_names, pol, theta, phi, widths, size, orders, rcwa_options):
    """
    Calculates the depth-dependent absorption profile using RCWA per wavelength. Called from calculate_absorption_profile_rcwa,
    can be used in parallel with joblib.

    :param wl: wavelength in nm
    :param A: 'A_pol' (polarization-dependent layer absorption) output from calculate_rat_rcwa
    :param geom_list: list containing shape information fpr each layer. Format is a list of lists; entries of inner lists are one dictionary
        per shape containing the shape information relevant for that shape.
    :param layers_oc: list of optical constants of the main layers in the stack
    :param shapes_oc: list of optical constants for the shapes inside the layers
    :param s_names: list of names of the shapes
    :param pol: polarization ('s', 'p' or 'u')
    :param theta: polar incidence angle in degrees
    :param phi: azimuthal angle (clockwise from y-axis) in degrees)
    :param widths: layer widths in nm. Incidence and transmission medium have 0 thickness.
    :param size: a tuple of 2-D vectors in the format ((ux, uy), (vx, vy)) giving the x and y components of the lattice unit vectors in nm.
    :param orders: number of orders to retain in the RCWA calculations.
    :param rcwa_options: S4 options (dictionary)
    :return: profile data (absorbed energy density at each depth)
    """
    def vs_pol_prof(s, p, A_total):
        profile = np.zeros(len(dist))
        S.SetExcitationPlanewave((theta, phi), s, p, 0)
        S.SetFrequency(1 / wl)
        for j, d in enumerate(dist):
            layer, d_in_layer = tmm.find_in_structure_with_inf(widths, d)
            layer_name = 'layer_' + str(layer + 1)  # layer_1 is air above so need to add 1
            data = rcwa_position_resolved(S, layer_name, d_in_layer, A_total, theta, n_inc)
            profile[j] = data

        return profile

    S = initialise_S(size, orders, geom_list, layers_oc, shapes_oc, s_names, widths, rcwa_options)
    n_inc = np.real(np.sqrt(layers_oc[0]))
    if len(pol) == 2:

        profile_data = vs_pol_prof(pol[0], pol[1], A)

    else:
        if pol in 'sp':
            profile_data = vs_pol_prof(int(pol == "s"), int(pol == "p"), A)

        else:
            profile_s = vs_pol_prof(1, 0, A[0])
            profile_p = vs_pol_prof(0, 1, A[1])
            profile_data = 0.5*(profile_s + profile_p)

    profile_data = profile_data
    return profile_data



from .adachi_alpha import create_adachi_alpha
from .absorption_QW import calc_alpha, calc_emission
from .transfer_matrix import calculate_absorption_profile, calculate_rat, calculate_ellipsometry, OptiStack
from .dielectric_constant_models import DielectricConstantModel
from .sopra_db import sopra_database
from .nk_db import download_db, search_db, create_nk_txt
from .cppm import Custom_CPPB
from .tmm_core_vec import inc_tmm, position_resolved, inc_tmm, unpolarized_RT, ellips, find_in_structure


class RCWASolverError(Exception):
    pass


try:
    from .rigorous_coupled_wave import calculate_absorption_profile_rcwa, calculate_rat_rcwa, \
        rcwa_rat, initialise_S, necessary_materials, update_epsilon, rcwa_position_resolved
except RCWASolverError:
    print('WARNING: The RCWA solver will not be available because an S4 installation has not been found.')



from solcore.structure import Layer, Junction, TunnelJunction
from solcore.absorption_calculator import calculate_rat, OptiStack, calculate_absorption_profile

import numpy as np
import types
from warnings import warn


def solve_tmm(solar_cell, options):
    """ Calculates the reflection, transmission and absorption of a solar cell object using the transfer matrix method.
    Internally, it creates an OptiStack and then it calculates the optical properties of the whole structure.
    A substrate can be specified in the SolarCell object, which is treated as a semi-infinite transmission medium.
    Shading can also be specified (as a fraction).

    Relevant options are 'wl' (the wavelengths, in m), the incidence angle 'theta' (in degrees), the polarization 'pol' ('s',
    'p' or 'u'), 'position' (locations in m at which depth-dependent absorption is calculated), 'no_back_reflection' and 'BL_correction'.
    'no_back_reflection' sets whether reflections from the back surface are suppressed (if set to True, the default),
    or taken into account (if set to False).

    If 'BL_correction' is set to True, thick layers (thickness > 10*maximum wavelength) are treated incoherently using
    the Beer-Lambert law, to avoid the calculation of unphysical interference oscillations in the R/A/T spectra.

    A coherency_list option can be provided: this should have elements equal to the total number of layers (if a Junction
    contains multiple Layers, each should have its own entry in the coherency list). Each element is either 'c' for coherent
    treatment of that layer or 'i' for incoherent treatment.

    :param solar_cell: A SolarCell object
    :param options: Options for the solver
    :return: None
    """
    wl = options.wavelength
    BL_correction = options.BL_correction if 'BL_correction' in options.keys() else True
    theta = options.theta if 'theta' in options.keys() else 0 # angle IN DEGREES
    pol = options.pol if 'pol' in options.keys() else 'u'
    zero_threshold = options.get("zero_threshold", 1e-5)

    # We include the shadowing losses
    initial = (1 - solar_cell.shading) if hasattr(solar_cell, 'shading') else 1

    # Now we calculate the absorbed and transmitted light. We first get all the relevant parameters from the objects
    all_layers = []
    widths = []
    n_layers_junction = []

    for j, layer_object in enumerate(solar_cell):

        # Attenuation due to absorption in the AR coatings or any layer in the front that is not part of the junction
        if type(layer_object) is Layer:
            all_layers.append(layer_object)
            widths.append(layer_object.width)
            n_layers_junction.append(1)

        # For each junction, and layer within the junction, we get the absorption coefficient and the layer width.
        elif type(layer_object) in [TunnelJunction, Junction]:
            n_layers_junction.append(len(layer_object))
            for i, layer in enumerate(layer_object):
                all_layers.append(layer)
                widths.append(layer.width)

    no_back_reflection = options.no_back_reflection if 'no_back_reflection' in options.keys() else True

    # With all the information, we create the optical stack
    if 'no_back_reflexion' in options.keys():
        warn('The no_back_reflexion warning is deprecated. Use no_back_reflection instead.', FutureWarning)
        no_back_reflection = options.no_back_reflexion


    full_stack = OptiStack(all_layers, no_back_reflection=no_back_reflection, substrate=solar_cell.substrate, incidence=solar_cell.incidence)

    if 'coherency_list' in options.keys():
        coherency_list = options.coherency_list
        coherent = False
        assert len(coherency_list) == full_stack.num_layers, \
            'Error: The coherency list must have as many elements (now {}) as the ' \
            'number of layers (now {}).'.format(len(coherency_list), full_stack.num_layers)
    else:
        coherency_list = None
        coherent = True

    if BL_correction and any(widths > 10*np.max(wl)): # assume it's safe to ignore interference effects
            make_incoherent = np.where(np.array(widths) > 10*np.max(wl))[0]
            print('Treating layer(s) ' + str(make_incoherent).strip('[]') + ' incoherently')
            if not 'coherency_list' in options.keys():
                coherency_list = np.array(len(all_layers)*['c'])
                coherent = False
            else:
                coherency_list = np.array(coherency_list)
            coherency_list[make_incoherent] = 'i'
            coherency_list = coherency_list.tolist()

    position = options.position * 1e9
    profile_position = position[position < sum(full_stack.widths)]

    print('Calculating RAT...')
    RAT = calculate_rat(full_stack, wl * 1e9, angle=theta,
                        coherent=coherent, coherency_list=coherency_list, no_back_reflection=no_back_reflection,
                        pol=pol)

    print('Calculating absorption profile...')
    out = calculate_absorption_profile(full_stack, wl * 1e9, dist=profile_position,
                                       angle=theta, no_back_reflection=no_back_reflection,
                                       pol=pol, coherent=coherent,
                                       coherency_list=coherency_list, zero_threshold=zero_threshold, RAT_out=RAT)

    # With all this information, we are ready to calculate the differential absorption function
    diff_absorption, all_absorbed = calculate_absorption_tmm(out, initial)

    # Each building block (layer or junction) needs to have access to the absorbed light in its region.
    # We update each object with that information.
    layer = 0
    A_per_layer = np.array(RAT['A_per_layer'][1:-1]) # first entry is R, last entry is T
    for j in range(len(solar_cell)):

        solar_cell[j].layer_absorption = initial*np.sum(A_per_layer[layer:(layer+n_layers_junction[j])], axis = 0)

        solar_cell[j].diff_absorption = diff_absorption
        solar_cell[j].absorbed = types.MethodType(absorbed, solar_cell[j])

        layer = layer + n_layers_junction[j]

    solar_cell.reflected = RAT['R'] * initial
    solar_cell.absorbed = sum([solar_cell[x].layer_absorption for x in np.arange(len(solar_cell))])
    solar_cell.transmitted = initial - solar_cell.reflected - solar_cell.absorbed


def absorbed(self, z):
    out = self.diff_absorption(self.offset + z) * (z < self.width)

    return out.T


def calculate_absorption_tmm(tmm_out, initial=1):
    all_z = tmm_out['position'] * 1e-9
    all_abs = initial * tmm_out['absorption'] / 1e-9

    def diff_absorption(z):
        idx = all_z.searchsorted(z)
        idx = np.where(idx <= len(all_z) - 2, idx, len(all_z) - 2)
        try:
            z1 = all_z[idx]
            z2 = all_z[idx + 1]

            f = (z - z1) / (z2 - z1)

            out = f * all_abs[:, idx] + (1 - f) * all_abs[:, idx + 1]

        except IndexError:
            out = all_abs[:, idx]

        return out

    all_absorbed = np.trapz(diff_absorption(all_z), all_z)

    return diff_absorption, all_absorbed
